"""
Module with functions to read or download time series with observations from knmi.

function levels:
1. get_obs_list_from_extent: list of observations from extent
    2. get_obs_list_from_codes: list of observations from codes
        3. get_lizard_groundwater: get single obs object
            4. get_timeseries_tube: get timeseries for a tube
                5. get_timeseries_uuid: get timeseries for a uuid

"""

import logging
import math
import pathlib
import warnings
from concurrent.futures import ThreadPoolExecutor

import geopandas
import pandas as pd
import requests
from pyproj import Transformer
from shapely.geometry import Polygon
from tqdm import tqdm

logger = logging.getLogger(__name__)

# TODO:
# - check transformation from EPSG:28992 to WGS84 (elsewhere in hydropandas we use
#   another definition for EPSG:28992 that is provided in util.py)

# Generic Lizard API endpoint (with 'organisation' as placeholder, following the Lizard documentation
lizard_api_endpoint = "https://{organisation}.lizard.net/api/v4/"


def check_status_obs(metadata, timeseries):
    """Checks if a monitoring tube is still active.

    If there are no measurements in the last 180 days, the monitoring
    tube is considered inactive.

    Parameters
    ----------
    metadata : dict
        metadata of the monitoring tube
    timeseries : pandas DataFrame
        timeseries of the monitoring well

    Returns
    -------
    metadata DataFrame including the status of the monitoring well
    """
    if timeseries.empty:
        metadata["status"] = "no timeseries available"
        return metadata

    last_measurement_date = timeseries.last_valid_index()
    today = pd.to_datetime("today").normalize()

    if today - last_measurement_date < pd.Timedelta(days=180):
        metadata["status"] = "active"

    else:
        metadata["status"] = "inactive"

    return metadata


def extent_to_wgs84_polygon(extent):
    """Translates an extent (xmin, xmax, ymin, ymax) to a polygon with coordinate system
    WGS84.

    Parameters
    ----------
    extent : list or tuple
        extent in epsg 28992 within which the observations are collected.

    Returns
    -------
    polygon of the extent with coordinate system WGS84
    """
    transformer = Transformer.from_crs("EPSG:28992", "WGS84")

    lon_min, lat_min = transformer.transform(extent[0], extent[2])
    lon_max, lat_max = transformer.transform(extent[1], extent[3])

    poly_T = Polygon(
        [(lat_min, lon_min), (lat_max, lon_min), (lat_max, lon_max), (lat_min, lon_max)]
    )

    return poly_T


def translate_flag(timeseries):
    """Translates Vitens Lizard flags from integer to text.

    Parameters
    ----------
    timeseries : pandas.DataFrame
        timeseries of a monitoring well with flags

    Returns
    -------
    timeseries : pandas.DataFrame
        timeseries with translated quality flags
    """
    translate_dic = {
        0: "betrouwbaar",
        1: "betrouwbaar",
        2: "onbeslist",
        3: "onbeslist",
        4: "onbeslist",
        6: "onbetrouwbaar",
        7: "onbetrouwbaar",
        99: "ongevalideerd",
        -99: "verwijderd",
    }
    timeseries["flag_description"] = timeseries["flag"].replace(translate_dic)

    return timeseries


def get_metadata_mw_from_code(code, organisation="vitens", auth=None):
    """Extracts the Groundwater Station parameters from a monitoring well based on the
    code of the monitoring well.

    Parameters
    ----------
    code : str
        code of the monitoring well
    organisation : str
        organisation indicating URL endpoint, currently only "vitens" is officially supported.
    auth : tuple, optional
        authentication credentials for the API request, e.g.: ("__key__", your_api_key)

    Raises
    ------
    ValueError
        if code of the monitoring well is not known

    Returns
    -------
    groundwaterstation_metadata : dict
        dictionary with all available metadata of the monitoring well and its filters
    """
    base_url = lizard_api_endpoint.format(organisation=organisation)
    lizard_GWS_endpoint = f"{base_url}groundwaterstations/"
    url_groundwaterstation_code = f"{lizard_GWS_endpoint}?code={code}"

    r = requests.get(url_groundwaterstation_code, auth=auth, timeout=1200)
    r.raise_for_status()

    try:
        groundwaterstation_metadata = r.json()["results"][0]
    except IndexError:
        raise ValueError(f"Code {code} is invalid. Response: {r.json()}")

    return groundwaterstation_metadata


def _prepare_API_input(nr_pages, url_groundwater):
    """Get API data pages within the defined extent.

    Parameters
    ----------
    nr_pages : int
        number of the pages on which the information is stored
    url_groundwater : str
        location of the used API to extract the data

    Returns
    -------
    urls : list
        list of the page number and the corresponding url
    """
    urls = []
    for page in range(nr_pages):
        true_page = page + 1  # The real page number is attached to the import thread
        urls += [url_groundwater + "&page={}".format(true_page)]
    return urls


def _download(url, timeout=1800, auth=None):
    """Function to download the data from the API using the ThreadPoolExecutor.

    Parameters
    ----------
    url : str
        url of an API page
    timeout : int, optional
        number of seconds to wait before terminating request
    auth : tuple, optional
        authentication credentials for the API request, e.g.: ("__key__", your_api_key)

    Returns
    -------
    dictionary with timeseries data
    """
    try:
        r = requests.get(url=url, timeout=timeout, auth=auth)
        r.raise_for_status()
        data = r.json()["results"]
    except requests.exceptions.Timeout:
        logger.error(f"Timeout while requesting {url}. Please check your connection.")
        data = []

    return data


def _split_mw_tube_nr(code):
    """get the tube number from a code that consists of the name and the tube number.

    Parameters
    ----------
    code : str
        name + tube_nr. e.g. 'BUWP014-11' or 'BUWP014012'

    Returns
    -------
    monitoring well, tube_number (str, int)

    Notes
    -----
    The format of the name + tube_nr is not very consistent and this function may need
    further finetuning.
    """

    if code[-3:].isdigit():
        return code[:-3], int(code[-3:])
    else:
        # assume there is a '-' to split name and filter number
        tube_nr = code.split("-")[-1]
        return code.strip(f"-{tube_nr}"), int(tube_nr)


def _extract_timeseries_info_from_tube(mtd_tube, auth=None):
    """
    Extracts timeseries information (hand/diver UUIDs and types) from a tube/filter dict.

    Parameters
    ----------
    mtd_tube : dict
        Tube/filter metadata dictionary.
    auth : tuple, optional
        Authentication credentials for the API request, e.g.: ("__key__", your_api_key)

    Returns
    -------
    dict
        Dictionary with timeseries info and type.
    """
    info = {}
    if not mtd_tube["timeseries"]:
        info["timeseries_type"] = None
        return info
    for series in mtd_tube["timeseries"]:
        r = requests.get(series, auth=auth)
        r.raise_for_status()
        series_info = r.json()
        # Note: See Github issue #311 for an explanation of 'wns_string'
        wns_string = series_info["code"].split(":", 1)[0]
        if wns_string == "WNS9040.hand":
            info["uuid_hand"] = series_info["uuid"]
            info["start_hand"] = series_info["start"]
        elif wns_string == "WNS9040":
            info["uuid_diver"] = series_info["uuid"]
            info["start_diver"] = series_info["start"]
        elif wns_string == "WNS9040.val":
            info["uuid_diver_validated"] = series_info["uuid"]
            info["start_diver_validated"] = series_info["start"]
            info["end_diver_validated"] = series_info["end"]

    # Create string with all timeseries types
    ts_types = []

    if info.get("start_hand") is not None:
        ts_types.append("hand")
    if info.get("start_diver") is not None:
        ts_types.append("diver")
    if info.get("start_diver_validated") is not None:
        ts_types.append("diver_validated")
    info["timeseries_type"] = " + ".join(ts_types) if ts_types else None

    return info


def get_metadata_tube(metadata_mw, tube_nr, auth=None):
    """Extract the metadata for a specific tube from the monitoring well metadata.

    Parameters
    ----------
    metadata_mw : dict
        dictionary with all available metadata of the monitoring well and all its
        filters
    tube_nr : int or None
        select metadata from a specific tube number
    auth : tuple, optional
        authentication credentials for the API request, e.g.: ("__key__", your_api_key)

    Raises
    ------
    ValueError
        if code of the monitoring well is invalid.

    Returns
    -------
    dictionary with metadata of a specific tube

    .. warning::
       This function assumes that there is only one 'hand' timeseries and one 'diver'
       timeseries for each tube. This seems to comply with the Vitens use of Lizard.
    """

    # Set default tube number if not provided
    if tube_nr is None:
        tube_nr = 1

    # Prepare a base metadata dict from the monitoring well
    metadata = {
        "location": metadata_mw["name"],
        "ground_level": metadata_mw["surface_level"],
        "source": "lizard",
        "organisation": metadata_mw["organisation"],
        "unit": "m NAP",
        "metadata_available": True,
        "status": None,
    }

    # Searches for filters matching the requested tube number
    metadata_tube_list = []
    for metadata_tube in metadata_mw["filters"]:
        # check if name+filternr ends with three digits
        code, tbnr = _split_mw_tube_nr(metadata_tube["code"])
        if tbnr == tube_nr:
            metadata_tube_list.append(metadata_tube)

    # Handles cases with no, one or multiple tubes with the same code and tube number
    if len(metadata_tube_list) == 0:
        raise ValueError(f"{metadata_mw['name']} doesn't have a tube number {tube_nr}")
    elif len(metadata_tube_list) == 1:
        mtd_tube = metadata_tube_list[0]
    elif len(metadata_tube_list) > 1:
        # tube has probably been replaced, multiple tubes with the same code and tube nr
        # merge metadata from all tubes
        logger.info(
            f"there are {len(metadata_tube_list)} instances of {code} and tube "
            f"{tube_nr}, trying to merge all in one observation object"
        )
        mtd_tube = metadata_tube_list[0].copy()
        relevant_keys = {
            "top_level",
            "filter_top_level",
            "filter_bottom_level",
            "timeseries",
        }
        for metadata_tube in metadata_tube_list:
            for key in set(metadata_tube.keys()) & relevant_keys:
                # check if properties are always the same for a tube number
                val = metadata_tube[key]
                if key in ["top_level", "filter_top_level", "filter_bottom_level"]:
                    if val != mtd_tube[key]:
                        logger.warning(
                            f"multiple {key} values found ({val} & {mtd_tube[key]})"
                            f" for {code} and tube {tube_nr}, using {mtd_tube[key]}"
                        )
                # merge time series from all tubes with the same code and tube number
                elif key == "timeseries":
                    mtd_tube[key] += val

        mtd_tube["code"] = f"{code}{tube_nr}"

    # Updates metadata with tube-specific information (top level, screen levels, coords)
    metadata.update(
        {
            "tube_nr": tube_nr,
            "name": mtd_tube["code"].replace("-", ""),
            "tube_top": mtd_tube["top_level"],
            "screen_top": mtd_tube["filter_top_level"],
            "screen_bottom": mtd_tube["filter_bottom_level"],
        }
    )

    lon, lat, _ = metadata_mw["geometry"]["coordinates"]
    transformer = Transformer.from_crs("WGS84", "EPSG:28992")
    metadata["x"], metadata["y"] = transformer.transform(lat, lon)

    # Extracts timeseries information (hand/diver UUIDs and types)
    metadata.update(_extract_timeseries_info_from_tube(mtd_tube, auth))

    return metadata


def get_timeseries_uuid(
    uuid, tmin, tmax, page_size=100000, organisation="vitens", auth=None
):
    """
    Get the time series (hand or diver) using the uuid.

    ----------
    uuid : str
        Universally Unique Identifier of the tube and type of time series.
    tmin : str YYYY-m-d
        start of the observations, by default the entire serie is returned
    tmax : str YYYY-m-d
        end of the observations, by default the entire serie is returned
    page_size : int, optional
        Query parameter which can extend the response size. The default is 100000.
    organisation : str, optional
        organisation as used by Lizard, currently only "vitens" is officially supported.
    auth : tuple, optional
        authentication credentials for the API request, e.g.: ("__key__", your_api_key)

    Returns
    -------
    pd.DataFrame
        pandas DataFrame with the timeseries of the monitoring well
    """
    base_url = lizard_api_endpoint.format(organisation=organisation)
    url_timeseries = f"{base_url}timeseries/{uuid}"

    if tmin is not None:
        tmin = pd.to_datetime(tmin).isoformat("T")

    if tmax is not None:
        tmax = pd.to_datetime(tmax).isoformat("T")

    params = {"start": tmin, "end": tmax, "page_size": page_size}
    url = url_timeseries + "/events/"

    r = requests.get(url=url, params=params, auth=auth)
    time_series_events = r.json()["results"]
    time_series_df = pd.DataFrame(time_series_events)

    if time_series_df.empty:
        return pd.DataFrame()

    else:
        time_series_df = translate_flag(time_series_df)

        timeseries_sel = time_series_df.loc[:, ["time", "value", "flag", "comment"]]
        timeseries_sel["time"] = pd.to_datetime(
            timeseries_sel["time"], format="%Y-%m-%dT%H:%M:%SZ", errors="coerce"
        ) + pd.DateOffset(hours=1)

        timeseries_sel = timeseries_sel[~timeseries_sel["time"].isnull()]

        timeseries_sel.set_index("time", inplace=True)
        timeseries_sel.index.rename("peil_datum_tijd", inplace=True)
        # timeseries_sel.dropna(inplace=True)

    return timeseries_sel


def _filter_timeseries(ts_dict, datafilters):
    """
    Generic filter function for multiple timeseries.

    Parameters
    ----------
    ts_dict : dict
        Dictionary of timeseries DataFrames, e.g. {'hand': df1, 'diver': df2, ...}
    datafilters : list of str
        List of datafilter names as strings, e.g. ["remove_unvalidated_diver_values_when_validated_available", "remove_hand_during_diver_period"]

    Returns
    -------
    dict
        Filtered ts_dict.
    """
    # Define standard filters by name. Note that the order may be relevant (uppermost filter is applied first).
    standard_datafilters = {
        "remove_unvalidated_diver_values_when_validated_available": {
            "target": "diver",
            "action": "remove_before",
            "reference": "diver_validated",
            "how": "max",
        },
        "remove_hand_during_diver_period": {
            "target": "hand",
            "action": "remove_between",
            "reference": ["diver", "diver_validated"],
            "how": "range",
        },
    }

    # Convert string datafilters to dicts using standard_datafilters
    datafilter_dicts = []
    for f in datafilters:
        if isinstance(f, str):
            if f in standard_datafilters:
                datafilter_dicts.append(standard_datafilters[f])
            else:
                raise ValueError(f"Unknown filter name: {f}")
        else:
            raise ValueError(
                "Each filter must be a string referring to a standard filter."
            )

    ts_dict = {k: v.copy() for k, v in ts_dict.items()}

    for f in datafilter_dicts:
        targets = f["target"] if isinstance(f["target"], list) else [f["target"]]
        refs = f["reference"] if isinstance(f["reference"], list) else [f["reference"]]
        how = f.get("how", "range")
        on = f.get("on", None)
        for target in targets:
            df = ts_dict.get(target)
            if df is None or df.empty:
                continue
            mask = pd.Series(True, index=df.index)
            for ref in refs:
                ref_df = ts_dict.get(ref)
                if ref_df is None or ref_df.empty:
                    continue
                idx = ref_df.index if on is None else ref_df[on]
                if f["action"] == "remove_before":
                    val = idx.max() if how == "max" else idx.min()
                    mask &= df.index > val
                elif f["action"] == "remove_after":
                    val = idx.min() if how == "min" else idx.max()
                    mask &= df.index < val
                elif f["action"] == "remove_between":
                    mask &= ~df.index.to_series().between(idx.min(), idx.max())
                elif f["action"] == "keep_only":
                    mask &= df.index.to_series().between(idx.min(), idx.max())
                # Add more actions as needed
            ts_dict[target] = df[mask]
    return ts_dict


def get_timeseries_tube(
    tube_metadata,
    tmin,
    tmax,
    type_timeseries=None,  # deprecated argument
    which_timeseries=("hand", "diver"),  # new preferred argument
    datafilters=None,
    combine_method="merge",
    organisation="vitens",
    auth=None,
):
    """
    Extracts specified timeseries for a tube and combines them as requested.

    Parameters
    ----------
    tube_metadata : dict
        metadata of a tube
    tmin : str YYYY-m-d, optional
        start of the observations
    tmax : str YYYY-m-d, optional
        end of the observations
    type_timeseries : str, optional (deprecated)
        deprecated, use 'which_timeseries' and 'combine_method' instead.
    which_timeseries : tuple of str, optional
        Which timeseries to retrieve. Options: "hand", "diver", "diver_validated".
        Defaults to ("hand", "diver") (which should be correct for Vitens).
    datafilters : list of strings, optional
        Methods to filter the timeseries data.
        If None (default), all measurements will be shown.
        Currently implemented datafilter methods:
        "remove_unvalidated_diver_values_when_validated_available": Removes diver values before last date with validated diver.
        "remove_hand_during_diver_period": Removes hand measurements during periods where diver or diver_validated measurements are available.
    combine_method : str, optional
        "merge" (vertical stack with 'origin' column) or "columns" (side-by-side columns).
        If None, defaults to "merge".
    organisation : str, optional
        organisation as used by Lizard.
    auth : tuple, optional
        authentication credentials for the API request, e.g.: ("__key__", your_api_key)

    Returns
    -------
    measurements : pandas DataFrame
        timeseries of the monitoring well
    metadata_df : dict
        metadata of the monitoring well
    """
    # Deprecation warning for type_timeseries
    if type_timeseries is not None:
        warnings.warn(
            "the argument 'type_timeseries' is deprecated and will eventually be "
            "removed, please use the arguments 'which_timeseries' and 'combine_method'.",
            DeprecationWarning,
        )
        # Map old type_timeseries to which_timeseries and combine_method
        if type_timeseries == "combine":
            combine_method = "columns"
        elif type_timeseries == "merge":
            combine_method = "merge"
        else:
            which_timeseries = [type_timeseries]
            combine_method = "merge"

    if tube_metadata["timeseries_type"] is None:
        return pd.DataFrame(), tube_metadata

    # Fetch requested timeseries
    ts_dict = {}
    for ts_type in which_timeseries:
        uuid_key = f"uuid_{ts_type}"
        start_key = f"start_{ts_type}"
        if tube_metadata.get(start_key) is not None:
            ts = get_timeseries_uuid(
                tube_metadata.get(uuid_key),
                tmin,
                tmax,
                organisation=organisation,
                auth=auth,
            )
        else:
            ts = pd.DataFrame()

        ts_dict[ts_type] = ts

    # Filter
    if datafilters is not None:
        ts_dict_filtered = _filter_timeseries(ts_dict, datafilters)
    else:
        ts_dict_filtered = ts_dict

    # Combine as requested
    if combine_method == "columns":
        # Side-by-side
        if not ts_dict_filtered.get("hand", pd.DataFrame()).empty:
            ts_dict_filtered["hand"] = ts_dict_filtered["hand"].rename(
                columns={"value": "value_hand", "flag": "flag_hand"}
            )
        if not ts_dict_filtered.get("diver", pd.DataFrame()).empty:
            ts_dict_filtered["diver"] = ts_dict_filtered["diver"].rename(
                columns={"value": "value_diver", "flag": "flag_diver"}
            )
        if not ts_dict_filtered.get("diver_validated", pd.DataFrame()).empty:
            ts_dict_filtered["diver_validated"] = ts_dict_filtered[
                "diver_validated"
            ].rename(
                columns={
                    "value": "value_diver_validated",
                    "flag": "flag_diver_validated",
                }
            )

        dfs = [
            df
            for key in ["hand", "diver_validated", "diver"]
            if key in ts_dict_filtered and not ts_dict_filtered[key].empty
            for df in [ts_dict_filtered[key]]
        ]
        measurements = pd.concat(dfs, axis=1) if dfs else pd.DataFrame()
        # Only keep present expected columns
        expected_cols = [
            "value_hand",
            "value_diver_validated",
            "value_diver",
            "flag_hand",
            "flag_diver_validated",
            "flag_diver",
        ]
        present_cols = [col for col in expected_cols if col in measurements.columns]
        if not measurements.empty:
            measurements = measurements.loc[:, present_cols]
    elif combine_method == "merge":
        # Default: merge (vertical stack)
        dfs = []
        for key in which_timeseries:
            df = ts_dict_filtered.get(key)
            if df is not None and not df.empty:
                df = df.copy()
                df["origin"] = key
                dfs.append(df)
        measurements = pd.concat(dfs, axis=0).sort_index() if dfs else pd.DataFrame()
    else:
        raise ValueError(
            f'{combine_method=} not supported, choose between "merge" and "columns".'
        )

    return measurements, tube_metadata


def get_lizard_groundwater(
    code,
    tube_nr=None,
    tmin=None,
    tmax=None,
    type_timeseries=None,  # deprecated argument
    which_timeseries=("hand", "diver"),  # new preferred argument
    datafilters=None,
    combine_method="merge",
    only_metadata=False,
    organisation="vitens",
    auth=None,
):
    """Extracts the metadata and timeseries of an observation well from a LIZARD-API
    based on the code of a monitoring well.

    Parameters
    ----------
    code : str
        code of the measuring well, e.g. '27B-0444'
    tube_nr : int, optional
        select specific tube top
        Default selects tube_nr = 1
    tmin : str YYYY-m-d, optional
        start of the observations, by default the entire serie is returned
    tmax : str YYYY-m-d, optional
        end of the observations, by default the entire serie is returned
    type_timeseries : str, optional (deprecated)
        Old keyword, use which_timeseries instead.
    which_timeseries : tuple of str, optional
        Which timeseries to retrieve. Options: "hand", "diver", "diver_validated".
        Defaults to ("hand", "diver") (which should be correct for Vitens).
    datafilters : list of strings, optional
        Methods to filter the timeseries data.
        If None (default), all measurements will be shown.
        Currently implemented datafilter methods:
        "remove_unvalidated_diver_values_when_validated_available": Removes diver values before last date with validated diver.
        "remove_hand_during_diver_period": Removes hand measurements during periods where diver or diver_validated measurements are available.
    combine_method : str, optional
        "merge" (vertical stack with 'origin' column) or "columns" (side-by-side columns).
        If None, defaults to "merge".
    only_metadata : bool, optional
        if True only metadata is returned and no time series data. The
        default is False.
    organisation : str, optional
        organisation as used by Lizard, currently only "vitens" is officially supported.
    auth : tuple, optional
        authentication credentials for the API request, e.g.: ("__key__", your_api_key)

    Returns
    -------
    measurements : pd.DataFrame
        returns a DataFrame with metadata and timeseries
    tube_metadata : dict
        dictionary containing metadata
    """

    groundwaterstation_metadata = get_metadata_mw_from_code(
        code, organisation=organisation, auth=auth
    )

    tube_metadata = get_metadata_tube(groundwaterstation_metadata, tube_nr, auth=auth)

    if only_metadata:
        return pd.DataFrame(), tube_metadata

    measurements, tube_metadata = get_timeseries_tube(
        tube_metadata,
        tmin,
        tmax,
        type_timeseries=type_timeseries,
        which_timeseries=which_timeseries,
        datafilters=datafilters,
        combine_method=combine_method,
        organisation=organisation,
        auth=auth,
    )
    tube_metadata = check_status_obs(tube_metadata, measurements)

    return measurements, tube_metadata


def get_obs_list_from_codes(
    codes,
    ObsClass,
    tube_nr="all",
    tmin=None,
    tmax=None,
    type_timeseries=None,  # deprecated argument
    which_timeseries=("hand", "diver"),  # new preferred argument
    datafilters=None,
    combine_method="merge",
    only_metadata=False,
    organisation="vitens",
    auth=None,
):
    """Get all observations from a list of codes of the monitoring wells and a list of
    tube numbers.

    Parameters
    ----------
    codes : list of str or str
        codes of the monitoring wells
    ObsClass : type
        class of the observations, e.g. GroundwaterObs
    tube_nr : lst of str
        list of tube numbers of the monitoring wells that should be selected.
        By default 'all' available tubes are selected.
    tmin : str YYYY-m-d, optional
        start of the observations, by default the entire serie is returned
    tmax : str YYYY-m-d, optional
        end of the observations, by default the entire serie is returned
    type_timeseries : str, optional (deprecated)
        Old keyword, use which_timeseries instead.
    which_timeseries : tuple of str, optional
        Which timeseries to retrieve. Options: "hand", "diver", "diver_validated".
        Defaults to ("hand", "diver") (which should be correct for Vitens).
    datafilters : list of strings, optional
        Methods to filter the timeseries data.
        If None (default), all measurements will be shown.
        Currently implemented datafilter methods:
        "remove_unvalidated_diver_values_when_validated_available": Removes diver values before last date with validated diver.
        "remove_hand_during_diver_period": Removes hand measurements during periods where diver or diver_validated measurements are available.
    combine_method : str, optional
        "merge" (vertical stack with 'origin' column) or "columns" (side-by-side columns).
        If None, defaults to "merge".
    only_metadata : bool, optional
        if True only metadata is returned and no time series data. The
        default is False.
    organisation : str, optional
        organisation as used by Lizard, currently only "vitens" is officially supported.
    auth : tuple, optional
        authentication credentials for the API request, e.g.: ("__key__", your_api_key)

    Returns
    -------
    obs_list
        list of observations
    """

    if isinstance(codes, str):
        codes = [codes]

    if not hasattr(codes, "__iter__"):
        raise TypeError("argument 'codes' should be an iterable")

    obs_list = []
    for code in codes:
        groundwaterstation_metadata = get_metadata_mw_from_code(
            code, organisation=organisation, auth=auth
        )
        tubes = []
        if tube_nr == "all":
            for metadata_tube in groundwaterstation_metadata["filters"]:
                tnr = _split_mw_tube_nr(metadata_tube["code"])[-1]
                if tnr not in tubes:
                    logger.debug(f"get {code}-{tnr}")
                    o = ObsClass.from_lizard(
                        code,
                        tnr,
                        tmin,
                        tmax,
                        type_timeseries=type_timeseries,
                        which_timeseries=which_timeseries,
                        datafilters=datafilters,
                        combine_method=combine_method,
                        only_metadata=only_metadata,
                        organisation=organisation,
                        auth=auth,
                    )
                    obs_list.append(o)
                    tubes.append(tnr)

        else:
            o = ObsClass.from_lizard(
                code,
                tube_nr,
                tmin,
                tmax,
                which_timeseries=which_timeseries,
                datafilters=datafilters,
                combine_method=combine_method,
                only_metadata=only_metadata,
                organisation=organisation,
                auth=auth,
            )
            obs_list.append(o)

    return obs_list


def get_obs_list_from_extent(
    extent,
    ObsClass,
    tube_nr="all",
    tmin=None,
    tmax=None,
    type_timeseries=None,  # deprecated argument
    which_timeseries=("hand", "diver"),  # new preferred argument
    datafilters=None,
    combine_method="merge",
    only_metadata=False,
    page_size=100,
    nr_threads=10,
    organisation="vitens",
    auth=None,
):
    """Get all observations within a specified extent.

    Parameters
    ----------
    extent : list or shapefile
        get groundwater monitoring wells within this extent [xmin, xmax, ymin, ymax]
        or within a predefined Polygon from a shapefile
    ObsClass : type
        class of the observations, e.g. GroundwaterObs
    tube_nr : lst of str
        list of tube numbers of the monitoring wells that should be selected.
        By default 'all' available tubes are selected.
    tmin : str, optional
        start of the observations (format YYYY-m-d), by default the entire series
        is returned
    tmax : str, optional
        end of the observations (format YYYY-m-d), by default the entire series
        is returned
    type_timeseries : str, optional (deprecated)
        Old keyword, use which_timeseries instead.
    which_timeseries : tuple of str, optional
        Which timeseries to retrieve. Options: "hand", "diver", "diver_validated".
        Defaults to ("hand", "diver") (which should be correct for Vitens).
    datafilters : list of strings, optional
        Methods to filter the timeseries data.
        If None (default), all measurements will be shown.
        Currently implemented datafilter methods:
        "remove_unvalidated_diver_values_when_validated_available": Removes diver values before last date with validated diver.
        "remove_hand_during_diver_period": Removes hand measurements during periods where diver or diver_validated measurements are available.
    combine_method : str, optional
        "merge" (vertical stack with 'origin' column) or "columns" (side-by-side columns).
        If None, defaults to "merge".
    only_metadata : bool, optional
        if True only metadata is returned and no time series data. The
        default is False.
    organisation : str
        organisation as used by Lizard, currently only "vitens" is officially supported.
    auth : tuple, optional
        authentication credentials for the API request, e.g.: ("__key__", your_api_key)
    page_size : int, optional
        number of records to retrieve per page, default is 100
    nr_threads : int, optional
        number of threads to use for the API requests, default is 10


    Returns
    -------
    obs_col : TYPE
        ObsCollection DataFrame with the 'obs' column
    """

    if isinstance(extent, (list, tuple)):
        polygon_T = extent_to_wgs84_polygon(extent)

    elif isinstance(extent, str) or isinstance(extent, pathlib.PurePath):
        polygon = geopandas.read_file(extent)
        # TODO: check this transformation
        polygon_T = polygon.to_crs("WGS84", "EPSG:28992").loc[0, "geometry"]
    else:
        raise TypeError("Extent should be a shapefile or a list of coordinates")

    base_url = lizard_api_endpoint.format(organisation=organisation)
    lizard_GWS_endpoint = f"{base_url}groundwaterstations/"
    url_groundwaterstation_extent = (
        f"{lizard_GWS_endpoint}?geometry__within={polygon_T}&page_size={page_size}"
    )
    r = requests.get(url_groundwaterstation_extent, auth=auth)
    r.raise_for_status()
    groundwaterstation_data = r.json()
    nr_results = groundwaterstation_data["count"]
    nr_pages = math.ceil(nr_results / page_size)

    logger.info("Number of monitoring wells: {}".format(nr_results))
    logger.info("Number of pages: {}".format(nr_pages))

    if nr_results == 0:
        ValueError(r.json())
        logger.warning(
            "No monitoring wells found in the specified extent. "
            "Please check the extent or the organisation."
        )
        return []

    if nr_threads > nr_pages:
        nr_threads = nr_pages

    urls = _prepare_API_input(nr_pages, url_groundwaterstation_extent)

    # Prepare arguments for get_obs_list_from_codes
    kwargs = {
        "ObsClass": ObsClass,
        "tube_nr": tube_nr,
        "tmin": tmin,
        "tmax": tmax,
        "type_timeseries": type_timeseries,
        "which_timeseries": which_timeseries,
        "datafilters": datafilters,
        "combine_method": combine_method,
        "only_metadata": only_metadata,
        "organisation": organisation,
        "auth": auth,
    }

    codes = []
    with ThreadPoolExecutor(max_workers=nr_threads) as executor:
        for result in tqdm(
            executor.map(lambda url: _download(url, auth=auth), urls),
            total=nr_pages,
            desc="Page",
        ):
            codes += [d["code"] for d in result]

    obs_list = []
    with ThreadPoolExecutor() as executor:
        for obs_list_mw in tqdm(
            executor.map(lambda code: get_obs_list_from_codes(code, **kwargs), codes),
            total=len(codes),
            desc="monitoring well",
        ):
            obs_list += obs_list_mw

    return obs_list



# %%
# packages
import numpy as np
import pandas as pd
import requests

from hydropandas.io import knmi


class EmptyDataFrameError(Exception):
    pass


# %%
# meteo stations
meteo_df = pd.read_json("knmi_meteostation.json")
meteo_dft = meteo_df.copy()

didx = [pd.Timestamp.today().normalize() - pd.Timedelta(days=1)] + list(
    reversed(
        pd.date_range(pd.Timestamp("1900-01-01"), pd.Timestamp.today(), freq="10YS")
    )
)


for stn in meteo_dft.index:
    df_daily, _ = knmi.get_knmi_daily_meteo_url(stn)
    vars_available_daily = ~df_daily.isna().all().drop(index=["STN"])
    meteo_dft.loc[stn, vars_available_daily.index] = vars_available_daily.values

    for end, start in zip(didx[:-1], didx[1:]):
        try:
            df_hourly, _ = knmi.get_knmi_hourly_meteo_api(stn, start=start, end=end)
            if df_hourly.empty:
                raise EmptyDataFrameError
            vars_available_hourly = ~df_hourly.isna().all().drop(index=["STN"])
            vah = vars_available_hourly[vars_available_hourly]
            meteo_dft.loc[stn, vah.index] = vah.values
        except (
            pd.errors.EmptyDataError,
            requests.ConnectionError,
            EmptyDataFrameError,
        ):
            print(f"Geen data {stn}, {start}, {end}")

meteo_dft = meteo_dft.fillna(False).drop(columns=[""])

# add tmin, tmax
meteo_tminmax = (
    pd.read_html("https://www.knmi.nl/nederland-nu/klimatologie/daggegevens")[0]
    .dropna(how="any")
    .convert_dtypes()
    .set_index("Nummer")
)
meteo_tminmax["tmin"] = [
    pd.to_datetime(x, format="%Y%m%d").strftime("%Y-%m-%d")
    for x in meteo_tminmax["Vanaf"]
]
meteo_tminmax["tmax"] = [
    pd.to_datetime(x, format="%Y%m%d") for x in meteo_tminmax["Tot en met"]
]
meteo_tminmax["tmax"] = [
    None
    if (pd.Timestamp.today() - x) < pd.Timedelta(days=365)
    else x.strftime("%Y-%m-%d")
    for x in meteo_tminmax["tmax"]
]
meteo_dft.loc[meteo_tminmax.index, "tmin"] = meteo_tminmax["tmin"].values
meteo_dft.loc[meteo_tminmax.index, "tmax"] = meteo_tminmax["tmax"].values

meteo_dft.to_json("knmi_meteostation.json")
# %%
# neerslag stations
prec_df = pd.read_json("knmi_neerslagstation.json")
prec_dft = prec_df.copy()
prec_tminmax = (
    pd.concat(
        pd.read_html(
            "https://www.knmi.nl/nederland-nu/klimatologie/monv/reeksen", index_col=0
        )
    )
    .reset_index()
    .set_index("Nr")
)
unknown_location = prec_tminmax.index[~np.isin(prec_tminmax.index, prec_dft.index)]
prec_tminmax = prec_tminmax.drop(unknown_location)
tminmax = [[x[0], x[-1]] for x in prec_tminmax["Periode"].str.split(" ")]
tmin = [pd.to_datetime(x[0], format="%Y%m%d").strftime("%Y-%m-%d") for x in tminmax]
tmax = [pd.to_datetime(x[1], format="%Y%m%d") for x in tminmax]
tmax = [
    None
    if (pd.Timestamp.today() - x) < pd.Timedelta(days=365)
    else x.strftime("%Y-%m-%d")
    for x in tmax
]
prec_dft.loc[prec_tminmax.index, "tmin"] = tmin
prec_dft.loc[prec_tminmax.index, "tmax"] = tmax

# Sometimes there are no measurements for a period after tmin, this is not a problem
# for all station but De Bilt. This is why we manually correct De Bilt.
prec_dft.loc[550, "tmin"] = "1898-01-01"

prec_dft.sort_index().loc[
    :, ["lon", "lat", "name", "x", "y", "altitude", "tmin", "tmax", "RD"]
].to_json("knmi_neerslagstation.json")



import json
import logging
import zipfile
from io import BytesIO, StringIO
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import requests
from platformdirs import user_data_dir
from shapely.geometry import box
from tqdm import tqdm

logger = logging.getLogger(__name__)

DH_zip_url = (
    "https://www.waterconnect.sa.gov.au/Content/Downloads/DEW/WATER_Drillholes_shp.zip"
)


def get_obs_list_from_extent(
    extent,
    ObsClass,
    tmin=None,
    tmax=None,
    only_metadata=False,
    keep_all_obs=False,
    location_gdf=None,
    update=False,
    **kwargs,
):
    """Get observations within a specific extent and optionally for a specific set of
    locations.

    Parameters
    ----------
    extent : list, tuple, numpy-array or None, optional
        get water connect measurements within this extent
        [xmin, xmax, ymin, ymax]
    ObsClass : type
        class of the observations, e.g. GroundwaterObs
    tmin : str or None, optional
        start time of observations. The default is None.
    tmax : str or None, optional
        end time of observations. The default is None.
    only_metadata : bool, optional
        if True download only metadata, significantly faster. The default
        is False.
    keep_all_obs : bool, optional
        if False, only observations with measurements are kept. The default
        is True.
    location_gdf : GeoDataFrame, optional
        geodataframe with the locations of the water drill holes you want to include.
    update : bool, optional
        if True new locations are downloaded and stored locally (slow) otherwise a
        cached version of the locations is used. By default False
    **kwargs
        additional keyword arguments are passed to the ObsClass.from_waterconnect()
        method

    Returns
    -------
    obs_list : list
        list with Obs objects

    """

    if only_metadata and not keep_all_obs:
        logger.warning(
            "Option with only_metadata is True and keep_all_obs is False not supported"
            "setting keep_all_obs=True"
        )
        keep_all_obs = True

    if location_gdf is None:
        location_gdf = get_locations_gdf(update=update)

    if extent is not None:
        location_gdf = get_locations_within_extent(location_gdf, extent)

    if location_gdf.empty:
        msg = f"No water connect measurements found within extent {extent}"
        logger.warning(msg)
        return []

    nr_obs_wells = len(location_gdf)
    logger.info(f"downloading waterconnect data from {nr_obs_wells} observation points")

    obs_list = []
    for dh_no, meta_series in tqdm(
        location_gdf.iterrows(), total=nr_obs_wells, desc="obs well"
    ):
        o = ObsClass.from_waterconnect(
            dh_no,
            meta_series,
            tmin=tmin,
            tmax=tmax,
            only_metadata=only_metadata,
            **kwargs,
        )
        if o.empty and not keep_all_obs:
            continue

        obs_list.append(o)

    return obs_list


def get_locations_gdf(fdir=None, keep_cols="all", update=False):
    """get locations of water drillholes in Southern Australia from water connect.

    Parameters
    ----------
    fdir : Path or str
        directory to store geodataframe with locations. If None a directory is created.
        The default is None.
    keep_cols : list or str, optional
        if provided only these columns of the geodataframe are stored. If 'all' all the
        columns are returned, by default 'all'.
    update : bool, optional
        if True new locations are downloaded and stored locally (slow) otherwise a
        cached version of the locations is used. By default False

    Returns
    -------
    geopandas.GeoDataFrame
        locations of water drillholes in Southern Australia
    """

    # set directories
    if fdir is None:
        fdir = Path(user_data_dir("waterconnect", "hydropandas"))
    elif isinstance(str, fdir):
        fdir = Path(fdir)
    fdir.mkdir(parents=True, exist_ok=True)
    fname_feather = fdir / "WATER_Drillholes_GDA2020.feather"

    # read files from feather
    if fname_feather.exists() and not update:
        logger.debug(f"Reading Water Connect locations from {fname_feather}")
        gdf_sel = gpd.read_feather(fname_feather)
        return gdf_sel

    msg = (
        "Initializing Water Connect, this may take several minutes. This only has to "
        "be done the first time you use Water Connect via hydropandas."
    )
    logger.info(msg)

    # download data
    r = requests.get(DH_zip_url, timeout=3600)
    r.raise_for_status()

    # extract files from zip and write to disk
    zip_buffer = BytesIO(r.content)
    with zipfile.ZipFile(zip_buffer, "r") as z:
        target_files = [
            name for name in z.namelist() if name.startswith("WATER_Drillholes_GDA2020")
        ]
        # Extract only those files
        logger.debug(f"Writing Water Connect locations shape to {fdir}")
        for file in target_files:
            z.extract(file, fdir)

    # read shapefiles to geodataframe
    logger.debug(f"Reading Water Connect locations from {fdir}")
    gdf_gda2020 = gpd.read_file(fdir / "WATER_Drillholes_GDA2020.shp")

    # select columns
    if keep_cols is None:
        keep_cols = [
            "UNIT_NO",
            "DHNO",
            "NAME",
            "LAT",
            "LON",
            "REF_ELEV",
            "GRND_ELEV",
            "STATE",
            "STATUS",
            "STAT_DESC",
            "PURPOSE",
            "PURP_DESC",
            "DRILL_DATE",
            "ORIG_DEPTH",
            "MAX_DEPTH",
            "CASE_TO",
            "geometry",
        ]
    elif isinstance(keep_cols, str) and keep_cols == "all":
        keep_cols = gdf_gda2020.columns

    if "geometry" not in keep_cols:
        keep_cols = list(keep_cols)
        keep_cols.append("geometry")

    gdf_sel = gdf_gda2020[keep_cols]

    # set index column
    gdf_sel.set_index("DHNO", inplace=True)

    # write to feather
    logger.debug(f"Write Water Connect locations to feather {fname_feather}")
    gdf_sel.to_feather(fname_feather)

    return gdf_sel


def get_locations_within_extent(gdf, extent=None):
    """get drillhole locations within an extent

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        all drillhole locations in Southern Australia
    extent : list, tuple, np.array, optional
        the extent, by default None

    Returns
    -------
    geopandas.GeoDataFrame
        GeoDataFrame with locations in the given extent
    """

    if extent is None:
        return gdf

    pol_extent = box(*tuple(np.asarray(extent)[[0, 2, 1, 3]]))
    gdf_extent = gdf.loc[gdf.within(pol_extent)]
    return gdf_extent


def get_waterconnect_obs(
    dh_no,
    meta_series=None,
    tmin=None,
    tmax=None,
    only_metadata=False,
    verify=True,
    pumping=True,
    anomalous=True,
    **kwargs,
):
    """get waterconnect observations using the API

    Parameters
    ----------
    dh_no : int or str
        drill hole number
    meta_series : pd.Series, optional
        series with metadata. Typically a row of the locations gdf.
    tmin : str or None, optional
        start time of observations. The default is None.
    tmax : str or None, optional
        end time of observations. The default is None.
    only_metadata : bool, optional
        if True only metadata and no measurements are returned. BY default False
    verify : bool, optional
        use verification to get a secure connection
    pumping : bool, optional
        return observations from pumping wells
    anomalous : bool, optional
        return anomalous observations
    **kwargs
        kwargs are passed to 'get_location_gdf'

    Returns
    -------
    pd.DataFrame: measurements

    dict: metadata

    """
    if meta_series is None:
        gdf = get_locations_gdf(**kwargs)
        meta_series = gdf.loc[dh_no]
    elif isinstance(meta_series, pd.DataFrame):
        meta_series = gdf.loc[dh_no]

    if only_metadata:
        measurements = pd.DataFrame()
        source = "water connect"
        unit = ""
    else:
        logger.debug(f"reading measurements for {dh_no}")
        measurements = get_measurements(
            dh_no,
            tmin=tmin,
            tmax=tmax,
            verify=verify,
            pumping=pumping,
            anomalous=anomalous,
        )

        if measurements.empty:
            logger.debug(f"no measurements for {dh_no}")
            source = "water connect"
            unit = ""
        else:
            if "data_source" in measurements.columns:
                source = (
                    ", ".join(measurements["data_source"].unique()) + " (water connect)"
                )
            else:
                source = "water connect"
            unit = "m AHD"

    meta = {
        "name": dh_no,
        "x": meta_series.pop("LON"),
        "y": meta_series.pop("LAT"),
        "unit": unit,
        "location": meta_series.pop("UNIT_NO"),
        "ground_level": meta_series.pop("GRND_ELEV"),
        "source": source,
        "meta": meta_series.to_dict(),
    }

    return measurements, meta


def get_measurements(
    dh_no, tmin=None, tmax=None, verify=True, pumping=True, anomalous=True
):
    """get measurement from a dh_no using the water connect api.

    Parameters
    ----------
    dh_no : int or str
        drill hole number
    tmin : str or None, optional
        start time of observations. The default is None.
    tmax : str or None, optional
        end time of observations. The default is None.
    verify : bool, optional
        use verification to get a secure connection
    pumping : bool, optional
        return observations from pumping wells
    anomalous : bool, optional
        return anomalous observations

    Returns
    -------
    pd.DataFrame with measurements, an empty DataFrame is returned if there are no
    measurements
    """
    result_str = request_api(dh_no, verify=verify, pumping=pumping, anomalous=anomalous)
    df = pd.read_csv(result_str)

    if df.empty:
        return df

    df.index = pd.to_datetime(df.pop("obs_date"), format="%d/%m/%Y")
    df = df.loc[tmin:tmax]

    if "rswl" in df.columns:
        df.insert(0, "rswl", df.pop("rswl"))

    return df


def request_api(
    dh_no, verify=True, pumping=True, anomalous=True, fname=None
) -> StringIO:
    """Download water connect data from the API

    Parameters
    ----------
    dh_no : int or str
        drill hole number
    verify : bool, optional
        use verification to get a secure connection
    pumping : bool, optional
        return observations from pumping wells
    anomalous : bool, optional
        return anomalous observations
    fname : str or Path, optional
        write requested data to this filename. Only if fname is not None

    Returns
    -------
    StringIO

    """
    url = "https://www.waterconnect.sa.gov.au/_layouts/15/dfw.sharepoint.wdd/WDDDMS.ashx/GetWaterLevelDownload?bulkOutput=CSV"
    json_data = {"DHNOs": [dh_no], "Pumping": pumping, "Anomalous": anomalous}
    r = requests.post(
        url, verify=verify, data={"exportdata": json.dumps(json_data)}, timeout=60
    )
    r.raise_for_status()

    result_str = r.text

    if result_str.startswith("<!DOCTYPE html>"):
        raise RuntimeError("water connect API down\n" + result_str)

    if fname is not None:
        with open(fname, "w") as f:
            f.write(result_str)

    return StringIO(result_str)



import os
from pathlib import Path

import hydropandas as hpd
from hydropandas.io import fews

fews_testdata = Path(__file__).parent / "data/2019-FEWS-test"

dirname, fnames = hpd.util.get_files(
    str(fews_testdata / "WaalenBurg_201810-20190215_prod.zip"), ".xml"
)


def test_fews_highmemory() -> None:
    fews.read_xml_fname(
        os.path.join(dirname, fnames[0]), low_memory=False, ObsClass=hpd.WaterlvlObs
    )


def test_fews_lowmemory() -> None:
    fews.read_xml_fname(
        os.path.join(dirname, fnames[0]), low_memory=True, ObsClass=hpd.WaterlvlObs
    )


def test_obscollection_fews_selection() -> None:
    fews.read_xml_filelist(fnames, hpd.WaterlvlObs, dirname)


def test_fews_pid_wsvv() -> None:
    wsvv_pid = hpd.get_fews_pid("WSVV")
    oc = hpd.ObsCollection.from_fews_xml(
        str(fews_testdata / "test_wsvv_fews.xml"), ObsClass=wsvv_pid
    )
    assert isinstance(oc.iloc[0].obs, hpd.PrecipitationObs)
    assert isinstance(oc.iloc[1].obs, hpd.EvaporationObs)



"""Module with observation classes.

The Obs class is a subclass of a pandas DataFrame with
additional attributes and methods. The specific classes (GroundwaterObs,
WaterlvlObs, ...) are subclasses of the Obs class.

The subclasses of a dataframe can have additional attributes and methods.
Additional attributes have to be defined in the '_metadata' attribute. In order
to keep the subclass methods and attributes when selecting or slicing an object
you need the '_constructor' method.


More information about subclassing pandas DataFrames can be found here:
http://pandas.pydata.org/pandas-docs/stable/development/extending.html#extending-subclassing-pandas
"""

import logging
import numbers
import os
import warnings
from _io import StringIO
from typing import List, Optional

import numpy as np
import pandas as pd
from pandas._config import get_option
from pandas.api.types import is_numeric_dtype
from pandas.io.formats import console

logger = logging.getLogger(__name__)


class Obs(pd.DataFrame):
    """Generic class for a time series with measurements at a certain location.

    Unless specified explicitly the first numeric column in the observation is
    used for analysis and plotting.

    Parameters
    ----------
    name : str
        name
    x : int or float
        x coordinate of observation point
    y : int or float
        y coordinate of observation point
    location : str
        label for the location of the observation. Observations on the same location
        but with a different ObsType or tube number should have the same label.
    meta : dictionary
        metadata
    filename : str
        filename with data of observation point
    source : str
        source of the observation e.g. BRO or KNMI
    unit : str
        unit of the first numerical column in the observation
    """

    # temporary properties
    _internal_names = pd.DataFrame._internal_names + ["none"]
    _internal_names_set = set(_internal_names)

    # normal properties
    _metadata = ["name", "x", "y", "location", "meta", "filename", "source", "unit"]

    def __init__(self, *args, **kwargs):
        """Constructor of Obs class.

        *args must be input for the pandas.DataFrame constructor, **kwargs can be one of
        the attributes listed in _metadata or keyword arguments for the constructor of a
        pandas.DataFrame.
        """
        if (len(args) > 0) and isinstance(args[0], Obs):
            for key in args[0]._get_meta_attr():
                if (key in Obs._metadata) and (key not in kwargs):
                    kwargs[key] = getattr(args[0], key)

        self.name = kwargs.pop("name", "")
        self.x = kwargs.pop("x", np.nan)
        self.y = kwargs.pop("y", np.nan)
        self.location = kwargs.pop("location", "")
        self.meta = kwargs.pop("meta", {})
        self.filename = kwargs.pop("filename", "")
        self.source = kwargs.pop("source", "")
        self.unit = kwargs.pop("unit", "")

        super(Obs, self).__init__(*args, **kwargs)

    def __repr__(self) -> str:
        """Return a string representation for a particular Observation."""
        buf = StringIO("")

        buf.write(f"{type(self).__name__} {self.name}\n")

        # write metadata properties
        buf.write("-----metadata------\n")
        for att in self._get_meta_attr():
            if not att == "meta":
                buf.write(f"{att} : {getattr(self, att)} \n")
        buf.write("\n")

        if self._info_repr():
            self.info(buf=buf)
            return buf.getvalue()

        buf.write("-----time series------\n")
        max_rows = get_option("display.max_rows")
        min_rows = get_option("display.min_rows")
        max_cols = get_option("display.max_columns")
        max_colwidth = get_option("display.max_colwidth")
        show_dimensions = get_option("display.show_dimensions")
        if get_option("display.expand_frame_repr"):
            width, _ = console.get_console_size()
        else:
            width = None
        self.to_string(
            buf=buf,
            max_rows=max_rows,
            min_rows=min_rows,
            max_cols=max_cols,
            line_width=width,
            max_colwidth=max_colwidth,
            show_dimensions=show_dimensions,
        )

        return buf.getvalue()

    def _repr_html_(self, collapse=False):
        """Uses the pandas DataFrame html representation with the metadata prepended."""
        obs_type = f'<p style="color:#808080";>hydropandas.{type(self).__name__}</p>\n'

        metadata_dic = {key: getattr(self, key) for key in self._get_meta_attr()}
        metadata_dic.pop("meta")
        metadata_df = pd.DataFrame(
            columns=[metadata_dic.pop("name")],
            index=metadata_dic.keys(),
            data=metadata_dic.values(),
        )
        metadata = metadata_df._repr_html_()

        observations = super()._repr_html_()
        if collapse:
            collapse_button_meta = (
                '<button type="button" class="collapsible '
                'active" id="meta"><i class="arrow right">'
                "</i> Metadata</button>\n"
            )
            collapse_button_obs = (
                '<button type="button" class="collapsible '
                'active" id="obs"><i class="arrow right">'
                "</i> Observations</button>\n"
            )

            with open(
                os.path.join(os.path.dirname(__file__), "static/style.css"), "r"
            ) as fo:
                css_arrow = fo.read()

            metadata = metadata.replace(
                "<div>\n<style scoped>", '<div  style="display: none;">\n<style scoped>'
            )
            observations = observations.replace(
                "<div>\n<style scoped>", '<div  style="display: none;">\n<style scoped>'
            )

            with open(
                os.path.join(os.path.dirname(__file__), "static/js_collapse.html"), "r"
            ) as fo:
                js_collapse_button = fo.read()

            return (
                obs_type
                + css_arrow
                + collapse_button_meta
                + metadata
                + "<br>\n"
                + collapse_button_obs
                + observations
                + js_collapse_button
            )
        return obs_type + metadata + "<br>\n" + observations

    @property
    def _constructor(self):
        return Obs

    def _get_meta_attr(self, ignore=("monitoring_well",)):
        """Get metadata attributes excluding the ones in ignore.

        Parameters
        ----------
        ignore : tuple, optional
            attributes to ignore, by default ('monitoring_well',)

        Returns
        -------
        list
            list of metadata attributes
        """

        return [a for a in self._metadata if a not in ignore]

    def _get_first_numeric_col_name(self):
        """Get the first numeric column name of the observations.

        Returns
        -------
        col : str, int or None
            Column name. None if there are no numeric columns.
        """
        if self.empty:
            return None

        for col in self.columns:
            if is_numeric_dtype(self[col]):
                return col

        return None

    def copy(self, deep=True):
        """Create a copy of the observation.

        When ``deep=True`` (default), a new object will be created with a
        copy of the calling object's data and indices. Modifications to
        the data or indices of the copy will not be reflected in the
        original object (see notes below).

        When ``deep=False``, a new object will be created without copying
        the calling object's data or index (only references to the data
        and index are copied). Any changes to the data of the original
        will be reflected in the shallow copy (and vice versa).

        Parameters
        ----------
        deep : bool, default True
            Make a deep copy, including a copy of the data and the indices.
            With ``deep=False`` neither the indices nor the data are copied.

        Returns
        -------
        o : hydropandas.Obs
            copy of the observation.
        """

        if not deep:
            return super().copy(deep=deep)

        o = super().copy(deep=deep)

        # creat a copy of all mutable attributes
        for att in self._get_meta_attr():
            val = getattr(self, att)
            if isinstance(val, (list, dict)):
                setattr(o, att, val.copy())

        return o

    def to_collection_dict(self, include_meta=False):
        """Get dictionary with registered attributes and their values of an Obs object.

        This method can be used to create a dataframe from a collection
        of Obs objects.

        Parameters
        ----------
        include_meta : boolean, optional
            include the meta dictionary in the collection dictionary,
            default is false

        Returns
        -------
        d : dictionary
            dictionary with Obs information
        """

        attrs = self._get_meta_attr()
        if not include_meta:
            attrs.remove("meta")

        d = {}
        for att in attrs:
            d[att] = getattr(self, att)

        d["obs"] = self

        return d

    def merge_metadata(self, right, overlap="error"):
        """Merge the metadata of an Obs object with metadata from another Obs object.

        Parameters
        ----------
        right : dict
            dictionary with the metadata of an Obs object.
        overlap : str, optional
            How to deal with overlapping metadata with different values.
            Options are:
                error : Raise a ValueError
                use_left : Use the metadata from self
                use_right : Use the given metadata
            Default is 'error'.

        Raises
        ------
        ValueError
            if the metadata differs and overlap='error'.

        Returns
        -------
        new_metadata : dict
            metadata after merge.
        """
        new_metadata = {}
        same_metadata = True
        for key in self._get_meta_attr():
            v1 = getattr(self, key)
            v2 = right[key]
            try:
                if v1 != v2:
                    # check if both are nan
                    if isinstance(v1, numbers.Number) and isinstance(
                        v2, numbers.Number
                    ):
                        if np.isnan(v1) and np.isnan(v2):
                            continue
                    same_metadata = False
                    if overlap == "error":
                        raise ValueError(
                            f"left observation {key} differs from right observation"
                        )
                    elif overlap == "use_left":
                        logger.info(
                            f"left observation {key} differs from right "
                            "observation, use left"
                        )
                        new_metadata[key] = v1
                    elif overlap == "use_right":
                        logger.info(
                            f"left observation {key} differs from right "
                            "observation, use right"
                        )
                        new_metadata[key] = v2
                else:
                    new_metadata[key] = v1
            except TypeError:
                same_metadata = False
                if overlap == "error":
                    raise ValueError(
                        f"left observation {key} differs from right observation"
                    )
                elif overlap == "use_left":
                    logger.info(
                        f"left observation {key} differs from right "
                        "observation, use left"
                    )
                    new_metadata[key] = v1
                elif overlap == "use_right":
                    logger.info(
                        f"left observation {key} differs from right "
                        "observation, use right"
                    )
                    new_metadata[key] = v2
        if same_metadata:
            logger.info("left and right observation have the same metadata")

        return new_metadata

    def _merge_timeseries(self, right, overlap="error"):
        """Merge two timeseries.

        Parameters
        ----------
        right : hpd.observation.Obs
            Observation object.
        overlap : str, optional
            How to deal with overlapping timeseries with different values.
            Options are:
                error : Raise a ValueError
                use_left : use the part of the overlapping timeseries from self
                use_right : use the part of the overlapping timeseries from
                right
            Default is 'error'.

        Raises
        ------
        ValueError
            when the time series have different values on the same timestamp.

        Returns
        -------
        Observation object.
        """
        o = self.iloc[0:0, 0:0]

        # check if time series are the same
        if self.equals(right):
            logger.info("left and right observation have the same time series")
            return self

        logger.info("right observation has a different time series")
        logger.info("merge time series")

        # merge overlapping columns
        overlap_cols = list(set(self.columns) & set(right.columns))

        # get timeseries for overlapping columns and indices
        dup_ind_o = right.loc[right.index.isin(self.index)]
        if not dup_ind_o.empty:
            # check if overlapping timeseries have different values
            if not self.loc[dup_ind_o.index, overlap_cols].equals(
                dup_ind_o[overlap_cols]
            ):
                logger.warning(
                    f"timeseries of observation {right.name} overlap with "
                    "different values"
                )
                if overlap == "error":
                    raise ValueError(
                        "observations have different values for same time steps"
                    )
                elif overlap == "use_left":
                    dup_o = self.loc[dup_ind_o.index, overlap_cols]
                elif overlap == "use_right":
                    dup_o = dup_ind_o[overlap_cols]
            else:
                dup_o = self.loc[dup_ind_o.index, overlap_cols]
        else:
            dup_o = dup_ind_o[overlap_cols]

        # get unique observations from overlapping columns
        unique_o_left = self.loc[
            ~self.index.isin(right.index)
        ]  # get unique observations left
        unique_o_right = right.loc[
            ~right.index.isin(self.index)
        ]  # get unique observations right

        # merge unique observations from overlapping columns with duplicate observations
        # from overlapping columns
        dfcol = pd.concat(
            [dup_o, unique_o_left[overlap_cols], unique_o_right[overlap_cols]]
        )
        o = pd.concat([o, dfcol], axis=1)

        # merge non overlapping columns
        for col in right.columns:
            if col not in o.columns:
                o = pd.concat([o, right[[col]]], axis=1)
        for col in self.columns:
            if col not in o.columns:
                o = pd.concat([o, self[[col]]], axis=1)

        o.sort_index(inplace=True)

        return o

    def merge_observation(self, right, overlap="error", merge_metadata=True):
        """Merge with another observation of the same type.

        Parameters
        ----------
        right : hpd.observation.Obs
            Observation object.
        overlap : str, optional
            How to deal with overlapping timeseries or metadata with different
            values.
            Options are:
                error : Raise a ValueError
                use_left : use the part of the overlapping timeseries from self
                use_right : use the part of the overlapping timeseries from
                right
            Default is 'error'.
        merge_metadata : bool, optional
            If True the metadata of the two objects are merged. If there are
            any differences the overlap parameter is used to determine which
            metadata is used. If merge_metadata is False, the metadata of self
            is always used for the merged observation. The default is True.

        Raises
        ------
        TypeError
            when the observation types are not the same.
        ValueError
            when the time series have different values on the same date or
            different values for the same metadata.

        Returns
        -------
        Observation object.
        """

        if overlap not in ["error", "use_left", "use_right"]:
            raise ValueError(
                "invalid value for overlap, choose between error, use_left anduse_right"
            )

        # check observation type
        if not isinstance(right, type(self)):
            raise TypeError(
                f"observation left has a different type {type(self)} than"
                f"observation right {type(right)}"
            )

        # merge timeseries
        o = self._merge_timeseries(right, overlap=overlap)

        # merge metadata
        if merge_metadata:
            metadata = {key: getattr(right, key) for key in right._get_meta_attr()}
            new_metadata = self.merge_metadata(metadata, overlap=overlap)
        else:
            new_metadata = {key: getattr(self, key) for key in self._get_meta_attr()}

        for key, item in new_metadata.items():
            setattr(o, key, item)

        return o


class GroundwaterObs(Obs):
    """Class for groundwater quantity observations.

    Subclass of the Obs class. Has the following attributes:
    - tube_nr: 2 tubes at one piezometer should have a different 'tube_nr'.
    - screen_top: top op the filter in m above date (NAP)
    - screen_bottom: bottom of the filter in m above date (NAP)
    - ground_level: surface level in m above date (NAP) (maaiveld in Dutch)
    - tube_top: top of the tube in m above date (NAP)
    - metadata_available: boolean indicating if metadata is available for
      the measurement point.

    Note
    ----
    In hydropandas version 0.13.0 the 'monitoring_well' attribute was removed and
    replaced by the 'location' attribute

    """

    _metadata = Obs._metadata + [
        "tube_nr",
        "screen_top",
        "screen_bottom",
        "ground_level",
        "tube_top",
        "metadata_available",
    ]

    def __init__(self, *args, **kwargs):
        """Constructor for ObsCollection.

        Parameters
        ----------
        *args must be input for the pandas.DataFrame constructor
        **kwargs can be one of the attributes listed in _metadata or keyword arguments
        for the constructor of a pandas.DataFrame.
        """
        if len(args) > 0:
            if isinstance(args[0], Obs):
                for key in args[0]._get_meta_attr():
                    if (key in GroundwaterObs._metadata) and (key not in kwargs.keys()):
                        kwargs[key] = getattr(args[0], key)

        if "monitoring_well" in kwargs:
            self.monitoring_well = kwargs.pop("monitoring_well", "")
        self.tube_nr = kwargs.pop("tube_nr", "")
        self.ground_level = kwargs.pop("ground_level", np.nan)
        self.tube_top = kwargs.pop("tube_top", np.nan)
        self.screen_top = kwargs.pop("screen_top", np.nan)
        self.screen_bottom = kwargs.pop("screen_bottom", np.nan)
        self.metadata_available = kwargs.pop("metadata_available", np.nan)

        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return GroundwaterObs

    @property
    def monitoring_well(self):
        msg = "The 'monitoring_well' attribute is deprecated and will be removed in hydropandas version 0.14.0., please use the 'location' attribute instead."
        warnings.warn(msg, FutureWarning)
        return self.location

    @monitoring_well.setter
    def monitoring_well(self, value):
        self.location = value

    @classmethod
    def from_bro(
        cls,
        bro_id,
        tube_nr=None,
        tmin="1900-01-01",
        tmax="2040-01-01",
        to_wintertime=True,
        drop_duplicate_times=True,
        only_metadata=False,
    ):
        """Download BRO groundwater observations from the server.

        Parameters
        ----------
        bro_id : str
            can be a GLD id or GMW id. If a GMW id is given a tube number is
            required as well. e.g. 'GLD000000012893'.
        tube_nr : str or None, optional
            if the bro_id is a GMW object the tube number should be given.
        tmin : str or None, optional
            start date in format YYYY-MM-DD
        tmax : str or None, optional
            end date in format YYYY-MM-DD
        to_wintertime : bool, optional
            if True the time index is converted to Dutch winter time. The default
            is True.
        drop_duplicate_times : bool, optional
            if True rows with a duplicate time stamp are removed keeping only the
            first row. The default is True.
        only_metadata : bool, optional
            if True only metadata is returned and no time series data. The
            default is False

        Returns
        -------
        TYPE
            DESCRIPTION.
        """

        from .io import bro

        measurements, meta = bro.get_bro_groundwater(
            bro_id,
            tube_nr,
            tmin=tmin,
            tmax=tmax,
            to_wintertime=to_wintertime,
            drop_duplicate_times=drop_duplicate_times,
            only_metadata=only_metadata,
        )

        return cls(
            measurements,
            meta=meta,
            name=meta.pop("name"),
            x=meta.pop("x"),
            y=meta.pop("y"),
            location=meta.pop("location"),
            source=meta.pop("source"),
            unit=meta.pop("unit"),
            screen_bottom=meta.pop("screen_bottom"),
            screen_top=meta.pop("screen_top"),
            ground_level=meta.pop("ground_level"),
            metadata_available=meta.pop("metadata_available"),
            tube_nr=meta.pop("tube_nr"),
            tube_top=meta.pop("tube_top"),
        )

    @classmethod
    def from_lizard(
        cls,
        code,
        tube_nr=None,
        tmin=None,
        tmax=None,
        type_timeseries=None,  # deprecated argument
        which_timeseries=["hand", "diver"],  # new preferred argument
        datafilters=None,
        combine_method="merge",
        only_metadata=False,
        organisation="vitens",
        auth=None,
    ):
        """Extracts the metadata and timeseries of a observation well from a LIZARD-API
        based on the code of a monitoring well.

        Parameters
        ----------
        code : str
            code of the measuring well
        tube_nr : int, optional
            select specific tube top
            Default selects tube_nr = 1
        tmin : str YYYY-m-d, optional
            start of the observations, by default the entire serie is returned
        tmax : str YYYY-m-d, optional
            end of the observations, by default the entire serie is returned
        type_timeseries : str, optional (deprecated)
            Old keyword, use which_timeseries instead.
        which_timeseries : list of str, optional
            Which timeseries to retrieve. Options: "hand", "diver", "diver_validated".
            Defaults to ["hand", "diver"] (which should be correct for Vitens).
        datafilters : list of strings, optional
            Methods to filter the timeseries data.
            If None (default), all measurements will be shown.
            Currently implemented datafilter methods:
            "remove_unvalidated_diver_values_when_validated_available": Removes diver values before last date with validated diver.
            "remove_hand_during_diver_period": Removes hand measurements during periods where diver or diver_validated measurements are available.
        combine_method : str, optional
            "merge" (vertical stack with 'origin' column) or "columns" (side-by-side columns).
            If None, defaults to "merge".
        only_metadata : bool, optional
            if True only metadata is returned and no time series data. The
            default is False.
        organisation : str, optional
            organisation of the data. Currently only 'vitens' is officially supported.
        auth : tuple, optional
            authentication credentials for the API request, e.g.: ("__key__", your_api_key)

        Returns
        -------
        ObsCollection
            Returns a DataFrame with metadata and timeseries
        """

        from .io import lizard

        measurements, meta = lizard.get_lizard_groundwater(
            code,
            tube_nr,
            tmin,
            tmax,
            type_timeseries=type_timeseries,
            which_timeseries=which_timeseries,
            combine_method=combine_method,
            datafilters=datafilters,
            only_metadata=only_metadata,
            organisation=organisation,
            auth=auth,
        )
        return cls(
            measurements,
            name=meta.pop("name"),
            x=meta.pop("x"),
            y=meta.pop("y"),
            location=meta.pop("location"),
            source=meta.pop("source"),
            unit=meta.pop("unit"),
            screen_bottom=meta.pop("screen_bottom"),
            screen_top=meta.pop("screen_top"),
            ground_level=meta.pop("ground_level"),
            metadata_available=meta.pop("metadata_available"),
            tube_nr=meta.pop("tube_nr"),
            tube_top=meta.pop("tube_top"),
            meta=meta,
        )

    @classmethod
    def from_bronhouderportaal_bro(
        cls,
        path,
        tube_nr,
        full_meta=False,
    ):
        """Load BRO groundwater metadata from XML file. Mind that bro_id is applicable,
        because file is not yet imported in BRO.

        Parameters
        ----------
        path : str
            filepath of XML file.
        tube_nr : int
            tube number.
        full_meta : bool
            process not only the standard metadata to ObsCollection.

        Returns
        -------
        ObsCollection
            ObsCollection containing observations from XML file.
        """

        from .io import bronhouderportaal_bro

        meta = bronhouderportaal_bro.get_metadata_from_gmw(
            path, tube_nr, full_meta=full_meta
        )

        empty_df = pd.DataFrame()

        return cls(
            empty_df,
            name=meta.pop("name"),
            x=meta.pop("x"),
            y=meta.pop("y"),
            location=meta.pop("location"),
            filename=meta.pop("filename"),
            source=meta.pop("source"),
            unit=meta.pop("unit"),
            screen_bottom=meta.pop("screen_bottom"),
            screen_top=meta.pop("screen_top"),
            ground_level=meta.pop("ground_level"),
            metadata_available=meta.pop("metadata_available"),
            tube_nr=meta.pop("tube_nr"),
            tube_top=meta.pop("tube_top"),
            meta=meta,
        )

    @classmethod
    def from_dino(
        cls,
        path=None,
        **kwargs,
    ):
        """Download dino data from the server.

        Parameters
        ----------
        path : str, optional
            path of dino csv file
        kwargs : key-word arguments
            these arguments are passed to hydropandas.io.dino.read_dino_groundwater_csv
            if path is not None and otherwise to hydropandas.io.dino.findMeetreeks
        """
        from .io import dino

        # read dino csv file
        measurements, meta = dino.read_dino_groundwater_csv(path, **kwargs)

        return cls(measurements, meta=meta, **meta)

    @classmethod
    def from_artdino_file(cls, path=None, **kwargs):
        """Read a dino csv file (artdiver style).

        Parameters
        ----------
        path : str, optional
            path of dino csv filename
        kwargs : key-word arguments
            these arguments are passed to hydropandas.io._dino.read_dino_groundwater_csv
        """
        from .io import dino

        # read artdino csv file
        measurements, meta = dino.read_artdino_groundwater_csv(path, **kwargs)

        return cls(measurements, meta=meta, **meta)

    @classmethod
    def from_waterconnect(
        cls,
        dh_no,
        meta_series=None,
        tmin=None,
        tmax=None,
        only_metadata=False,
        verify=True,
        pumping=True,
        anomalous=True,
        **kwargs,
    ):
        """Read data from water connect api.

        Parameters
        ----------
        dh_no : int or str
            drill hole number
        meta_series : pd.Series, optional
            series with metadata. Typically a row of the locations gdf.
        tmin : str or None, optional
            start time of observations. The default is None.
        tmax : str or None, optional
            end time of observations. The default is None.
        only_metadata : bool, optional
            if True only metadata and no measurements are returned. BY default False
        verify : bool, optional
            use verification to get a secure connection
        pumping : bool, optional
            return observations from pumping wells
        anomalous : bool, optional
            return anomalous observations
        **kwargs
            kwargs are passed to 'get_waterconnect_obs'

        Returns
        -------
        GroundwaterObs
            GroundwaterObs object

        Raises
        ------
        ValueError
            if file contains data for more than one location
        """
        from .io import water_connect

        df, metadata = water_connect.get_waterconnect_obs(
            dh_no,
            meta_series=meta_series,
            tmin=tmin,
            tmax=tmax,
            only_metadata=only_metadata,
            verify=verify,
            pumping=pumping,
            anomalous=anomalous,
            **kwargs,
        )

        return cls(df, **metadata)

    @classmethod
    def from_wiski(cls, path, **kwargs):
        """Read data from a WISKI file.

        Parameters:
        -----------
        path : str
            The path of the file to be read.
        sep : str, optional (default=";")
            The delimiter used to separate fields in the file.
        header_sep : str, optional (default=None)
            The delimiter used to separate fields in the header. If None, the
            function will try to automatically detect the separator.
        header_identifier : str, optional (default="#")
            The character used to identify header lines.
        read_series : bool, optional (default=True)
            Whether to read the time series data from the file.
        translate_dic : dict, optional (default=None)
            A dictionary mapping header field names to the desired output names.
        tz_localize : bool, optional (default=True)
            Whether to localize the datetime index to the machine's timezone.
        unit : str, optional (default="")
            The unit of measurement of the data.
        **kwargs : keyword arguments
            Additional arguments to pass to the pandas `read_csv` function.
        """
        from .io import wiski

        data, metadata = wiski.read_wiski_file(path, **kwargs)

        return cls(data, meta=metadata, **metadata)

    @classmethod
    def from_pastastore(cls, pstore, libname, name, metadata_mapping=None):
        """Read item from pastastore library.

        Parameters
        ----------
        pstore : pastastore.PastaStore
            pastastore object
        libname : str
            name of library containinig item
        name : str
            name of item
        metadata_mapping : dict, optional
            dictionary containing map between metadata field names in pastastore (keys)
            and metadata field names expected by hydropandas (values), by default None.
        """
        from .io import pastas

        data, metadata = pastas.read_pastastore_item(pstore, libname, name)

        if metadata_mapping is not None:
            for pstore_name, oc_name in metadata_mapping.items():
                metadata[oc_name] = metadata.get(pstore_name, None)

        metadata["source"] = "pastastore"
        kwargs = {}
        for key, value in metadata.items():
            if key in cls._metadata:
                kwargs[key] = value

        return cls(data, meta=metadata, **kwargs)

    @classmethod
    def from_solinst(
        cls,
        path,
        transform_coords=True,
        screen_bottom=None,
        screen_top=None,
        ground_level=None,
        tube_nr=None,
        tube_top=None,
    ):
        """Read data from Solinst xle file.

        Parameters
        ----------
        path : str
            path to file (file can zip or xle)

        """
        from .io import solinst

        df, meta = solinst.read_solinst_file(path, transform_coords=transform_coords)

        return cls(
            df,
            meta=meta,
            name=meta.pop("name"),
            x=meta.pop("x"),
            y=meta.pop("y"),
            location=meta.pop("location"),
            filename=meta.pop("filename"),
            source=meta.pop("source"),
            unit=meta.pop("unit"),
            screen_bottom=screen_bottom,
            screen_top=screen_top,
            ground_level=ground_level,
            metadata_available=meta.pop("metadata_available"),
            tube_nr=tube_nr,
            tube_top=tube_top,
        )


class WaterQualityObs(Obs):
    """Class for water quality ((grond)watersamenstelling) point observations.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + [
        "tube_nr",
        "ground_level",
        "metadata_available",
    ]

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            if isinstance(args[0], Obs):
                for key in args[0]._get_meta_attr():
                    if (key in WaterQualityObs._metadata) and (
                        key not in kwargs.keys()
                    ):
                        kwargs[key] = getattr(args[0], key)

        if "monitoring_well" in kwargs:
            self.monitoring_well = kwargs.pop("monitoring_well", "")
        self.tube_nr = kwargs.pop("tube_nr", "")
        self.ground_level = kwargs.pop("ground_level", np.nan)
        self.metadata_available = kwargs.pop("metadata_available", np.nan)

        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return WaterQualityObs

    @property
    def monitoring_well(self):
        msg = "The 'monitoring_well' attribute is deprecated and will be removed in hydropandas version 0.14.0., please use the 'location' attribute instead."
        warnings.warn(msg, FutureWarning)
        return self.location

    @monitoring_well.setter
    def monitoring_well(self, value):
        self.location = value

    @classmethod
    def from_dino(cls, path, **kwargs):
        """Read dino file with groundwater quality data.

        Parameters
        ----------
        path : str
            path of dino txt filename
        kwargs : key-word arguments
            these arguments are passed to
            hydropandas.io.dino.read_dino_groundwater_quality_txt
        """
        from .io import dino

        measurements, meta = dino.read_dino_groundwater_quality_txt(path, **kwargs)

        return cls(measurements, meta=meta, **meta)

    @classmethod
    def from_waterinfo(
        cls,
        path=None,
        location_gdf=None,
        locatie=None,
        grootheid_code=None,
        groepering_code=None,
        parameter_code=None,
        tmin=None,
        tmax=None,
        **kwargs,
    ):
        """Read data from waterinfo csv-file or zip.

        Parameters
        ----------
        path : str, optional
            path to file (file can zip or csv)
        location_gdf : geopandas.GeoDataFrame, optional
            geodataframe with locations, only used if path is None, default is None
        locatie : str or list of str, optional
            select only measurement with this location(s), e.g. 'SCHOONHVN', default is None
        grootheid_code : str or list of str, optional
            select only measurement with this grootheid_code, e.g. 'WATHTE', default is None
        groepering_code : str or list of str, optional
            select only measurement with this groepering_code, e.g. 'GETETBRKD2', default is None
        parameter_code :  str or list of str, optional
            select only measurement with this parameter_code, e.g. 'Cl', default is None
        tmin : pd.Timestamp or str, optional
            start date of the measurements, only used if path is None, default is None
        tmax : pd.Timestamp or str, optional
            end date of the measurements, only used if path is None, default is None

        Returns
        -------
        WaterlvlObs
            WaterlvlObs object

        Raises
        ------
        ValueError
            if file contains data for more than one location
        """
        from .io import waterinfo

        df, metadata = waterinfo.get_waterinfo_obs(
            path=path,
            location_gdf=location_gdf,
            locatie=locatie,
            grootheid_code=grootheid_code,
            groepering_code=groepering_code,
            parameter_code=parameter_code,
            tmin=tmin,
            tmax=tmax,
            **kwargs,
        )

        return cls(df, **metadata)


class WaterlvlObs(Obs):
    """Class for water level point observations.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + ["metadata_available"]

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            if isinstance(args[0], Obs):
                for key in args[0]._get_meta_attr():
                    if (key in WaterlvlObs._metadata) and (key not in kwargs.keys()):
                        kwargs[key] = getattr(args[0], key)

        if "monitoring_well" in kwargs:
            self.monitoring_well = kwargs.pop("monitoring_well", "")
        self.metadata_available = kwargs.pop("metadata_available", np.nan)

        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return WaterlvlObs

    @property
    def monitoring_well(self):
        msg = "The 'monitoring_well' attribute is deprecated and will be removed in hydropandas version 0.14.0., please use the 'location' attribute instead."
        warnings.warn(msg, FutureWarning)
        return self.location

    @monitoring_well.setter
    def monitoring_well(self, value):
        self.location = value

    @classmethod
    def from_dino(cls, path, **kwargs):
        """Read a dino file with waterlvl data.

        Parameters
        ----------
        path : str
            path of dino csv filename
        kwargs : key-word arguments
            these arguments are passed to hydropandas.io.dino.read_dino_waterlvl_csv
        """
        from .io import dino

        measurements, meta = dino.read_dino_waterlvl_csv(path, **kwargs)

        return cls(measurements, meta=meta, **meta)

    @classmethod
    def from_matroos(
        cls, location, source, unit, tmin=None, tmax=None, only_metadata=False, **kwargs
    ):
        """Read data using the Matroos API

        Parameters
        ----------
        location : str
            location e.g. 'krimpen a/d lek'
        unit : str
            unit e.g. 'waterlevel'
        source : str
            source e.g. 'observed'
        tmin : pd.Timestamp, str or None, optional
            start of time series if None tmin is 10 days ago, by default None
        tmax : pd.Timestamp, str or None, optional
            start of time series if None tmin is today, by default None
        only_metadata : bool, optional
            if True download only metadata, slightly faster. The default
            is False.

        Returns
        -------
        df : WaterlvlObs
            WaterlvlObs object

        Raises
        ------
        ValueError
            if file contains data for more than one location
        """
        from .io import matroos

        df, metadata = matroos.get_matroos_obs(
            location, unit, source, tmin=None, tmax=None, only_metadata=False, **kwargs
        )

        return cls(
            df,
            name=metadata.pop("name"),
            x=metadata.pop("x"),
            y=metadata.pop("y"),
            location=metadata.pop("location"),
            source=metadata.pop("source"),
            meta=metadata,
        )

    @classmethod
    def from_waterinfo(
        cls,
        path=None,
        location_gdf=None,
        locatie=None,
        grootheid_code=None,
        groepering_code=None,
        parameter_code=None,
        tmin=None,
        tmax=None,
        **kwargs,
    ):
        """Read data from waterinfo csv-file or zip.

        Parameters
        ----------
        path : str, optional
            path to file (file can zip or csv)
        location_gdf : geopandas.GeoDataFrame, optional
            geodataframe with locations, only used if path is None, default is None
        locatie : str or list of str, optional
            select only measurement with this location(s), e.g. 'SCHOONHVN', default is None
        grootheid_code : str or list of str, optional
            select only measurement with this grootheid_code, e.g. 'WATHTE', default is None
        groepering_code : str or list of str, optional
            select only measurement with this groepering_code, e.g. 'GETETBRKD2', default is None
        parameter_code :  str or list of str, optional
            select only measurement with this parameter_code, e.g. 'Cl', default is None
        tmin : pd.Timestamp or str, optional
            start date of the measurements, only used if path is None, default is None
        tmax : pd.Timestamp or str, optional
            end date of the measurements, only used if path is None, default is None

        Returns
        -------
        df : WaterlvlObs
            WaterlvlObs object

        Raises
        ------
        ValueError
            if file contains data for more than one location
        """
        from .io import waterinfo

        df, metadata = waterinfo.get_waterinfo_obs(
            path=path,
            location_gdf=location_gdf,
            locatie=locatie,
            grootheid_code=grootheid_code,
            groepering_code=groepering_code,
            parameter_code=parameter_code,
            tmin=tmin,
            tmax=tmax,
            **kwargs,
        )

        return cls(df, **metadata)


class ModelObs(Obs):
    """Class for model point results.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + ["model"]

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            if isinstance(args[0], Obs):
                for key in args[0]._get_meta_attr():
                    if (key in ModelObs._metadata) and (key not in kwargs.keys()):
                        kwargs[key] = getattr(args[0], key)

        self.model = kwargs.pop("model", "")

        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return ModelObs


class MeteoObs(Obs):
    """Class for meteorological timeseries.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + ["station", "meteo_var"]

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            if isinstance(args[0], Obs):
                for key in args[0]._get_meta_attr():
                    if (key in MeteoObs._metadata) and (key not in kwargs.keys()):
                        kwargs[key] = getattr(args[0], key)

        self.station = kwargs.pop("station", np.nan)
        self.meteo_var = kwargs.pop("meteo_var", "")

        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return MeteoObs

    @classmethod
    def from_knmi(
        cls,
        meteo_var,
        stn=None,
        fname=None,
        xy=None,
        start=None,
        end=None,
        fill_missing_obs=False,
        interval="daily",
        use_api=True,
        raise_exceptions=True,
        startdate=None,
        enddate=None,
    ):
        """Get a MeteoObs timeseries from the KNMI meteo data.

        Parameters
        ----------
        meteo_var : str
            meteo variable e.g. "RH" or "EV24". For a list of possible
            variables see the hydropandas.read_knmi function.
        stn : int, str or None, optional
            measurement station e.g. 829. The default is None.
        fname : str, path object, file-like object or None, optional
            filename of a knmi file. The default is None.
        xy : list, tuple or None, optional
            RD coördinates of a location in the Netherlands. The station nearest
            to this location used. The Default is None.
        start : str, datetime or None, optional
            start date of observations. The default is None.
        end : str, datetime or None, optional
            end date of observations. The default is None.
        fill_missing_obs : bool, optional
            if True nan values in time series are filled with nearby time series.
            The default is False. Note: if the given stn has no data between start and
            end the data from nearby stations is used. In this case the metadata of the
            Observation is the metadata from the nearest station that has any
            measurement in the given period.
        interval : str, optional
            desired time interval for observations. Options are 'daily' and
            'hourly'. The default is 'daily'.
        use_api : bool, optional
            if True the api is used to obtain the data, API documentation is here:
                https://www.knmi.nl/kennis-en-datacentrum/achtergrond/data-ophalen-vanuit-een-script
            if False a text file is downloaded into a temporary directory and the
            data is read from there. Default is True since the api is back
            online (July 2021).
        raise_exceptions : bool, optional
            if True you get errors when no data is returned. The default is False.

        Returns
        -------
        MeteoObs object with meteorological observations
        """
        from .io import knmi

        if startdate is not None:
            warnings.warn(
                "please use 'start' instead of startdate'", DeprecationWarning
            )
            start = startdate
        if enddate is not None:
            warnings.warn("please use 'end' instead of enddate'", DeprecationWarning)
            end = enddate

        ts, meta = knmi.get_knmi_obs(
            stn=stn,
            fname=fname,
            xy=xy,
            meteo_var=meteo_var,
            start=start,
            end=end,
            fill_missing_obs=fill_missing_obs,
            interval=interval,
            use_api=use_api,
            raise_exceptions=raise_exceptions,
        )

        return cls(
            ts,
            meta=meta,
            station=meta.pop("station"),
            x=meta.pop("x"),
            y=meta.pop("y"),
            name=meta.pop("name"),
            source=meta.pop("source"),
            unit=meta.pop("unit") if "unit" in meta else "",
            meteo_var=meteo_var,
        )

    @classmethod
    def from_wow(
        cls,
        meteo_var: str,
        stn: str = None,
        xy: List[float] = None,
        start: Optional[pd.Timestamp] = None,
        end: Optional[pd.Timestamp] = None,
    ):
        """Get a MeteoObs timeseries from a wow.knmi.nl station.

        Parameters
        ----------
        meteo_var : str
            wow meteo variable
        stn : Optional[int, str], optional
            station name
        xy : Optinal[List[float]], optinial
            longitude latitude of location [lon, lat] eg: [4.85, 51.95]
        start : Optional[pd.Timestamp], optional
            start date of observations, by default None
        end : Optional[pd.Timestamp], optional
            start date of observations, by default None

        Returns
        -------
        MeteoObs
        """
        from .io import wow

        wow_df, meta = wow.get_wow(
            stn=stn, xy=xy, meteo_var=meteo_var, start=start, end=end
        )

        return cls(
            wow_df,
            meta=meta,
            station=meta["site"]["id"],
            x=meta["site"]["geo"]["coordinates"][1],
            y=meta["site"]["geo"]["coordinates"][0],
            name=meta["site"]["name"],
            source="wow.knmi.nl",
            unit=wow_df.columns[0].split(" ")[-1],
            meteo_var=meteo_var,
        )


class EvaporationObs(MeteoObs):
    """Class for evaporation timeseries.

    Subclass of the MeteoObs class
    """

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            if isinstance(args[0], Obs):
                for key in args[0]._get_meta_attr():
                    if (key in EvaporationObs._metadata) and (key not in kwargs.keys()):
                        kwargs[key] = getattr(args[0], key)

        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return EvaporationObs

    @classmethod
    def from_knmi(
        cls,
        meteo_var="EV24",
        stn=None,
        fname=None,
        xy=None,
        start=None,
        end=None,
        fill_missing_obs=False,
        interval="daily",
        use_api=True,
        raise_exceptions=True,
        startdate=None,
        enddate=None,
    ):
        """Get an EvaporationObs timeseries from the KNMI evaporation in m.

        Parameters
        ----------
        meteo_var : str, optional
            meteo variable should be "EV24".
        stn : int, str or None, optional
            measurement station e.g. 829. The default is None.
        fname : str, path object, file-like object or None, optional
            filename of a knmi file. The default is None.
        xy : list, tuple or None, optional
            RD coördinates of a location in the Netherlands. The station nearest
            to this location used. The Default is None.
        start : str, datetime or None, optional
            start date of observations. The default is None.
        end : str, datetime or None, optional
            end date of observations. The default is None.
        fill_missing_obs : bool, optional
            if True nan values in time series are filled with nearby time series.
            The default is False.
        interval : str, optional
            desired time interval for observations. Options are 'daily' and
            'hourly'. The default is 'daily'.
        inseason : boolean, optional
            flag to obtain inseason data. The default is False
        raise_exceptions : bool, optional
            if True you get errors when no data is returned. The default is False.
        use_api : bool, optional
            if True the api is used to obtain the data, API documentation is here:
                https://www.knmi.nl/kennis-en-datacentrum/achtergrond/data-ophalen-vanuit-een-script
            if False a text file is downloaded into a temporary directory and the
            data is read from there. Default is True since the api is back
            online (July 2021).


        Returns
        -------
        EvaporationObs object with an evaporation time series and attributes
        """

        return super().from_knmi(
            meteo_var,
            stn=stn,
            fname=fname,
            xy=xy,
            start=start,
            end=end,
            fill_missing_obs=fill_missing_obs,
            interval=interval,
            use_api=use_api,
            raise_exceptions=raise_exceptions,
            startdate=startdate,
            enddate=enddate,
        )


class PrecipitationObs(MeteoObs):
    """Class for precipitation timeseries.

    Subclass of the MeteoObs class
    """

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            if isinstance(args[0], Obs):
                for key in args[0]._get_meta_attr():
                    if (key in PrecipitationObs._metadata) and (
                        key not in kwargs.keys()
                    ):
                        kwargs[key] = getattr(args[0], key)

        super().__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return PrecipitationObs

    @classmethod
    def from_knmi(
        cls,
        meteo_var="RH",
        stn=None,
        fname=None,
        xy=None,
        start=None,
        end=None,
        fill_missing_obs=False,
        interval="daily",
        use_api=True,
        raise_exceptions=True,
        startdate=None,
        enddate=None,
    ):
        """Get a PrecipitationObs timeseries from the KNMI precipitation. The
        precipitation is the Daily precipitation amount (in 0.1 mm) (-1 for.

        <0.05 mm).

        There are 3 different ways to obtain precipitation data from the knmi:
            1. Daily data from precipitation (neerslag) stations
            2. Daily data from meteo stations
            3. Hourly data from meteo stations

        1. For daily data from a precipitation station (neerslagstation) meteo_var
        should be 'RD'.
        2. For daily data from a meteo station meteo_var should be 'RH' and
        interval should be 'daily' (default).
        3. For hourly data from a meteo station meteo_var should be 'RH' and
        interval should be 'hourly'.

        More information about the differences between neerslag and meteo
        stations can be found in the hydropandas documentation ->
        02_knmi_observations notebook.

        Parameters
        ----------
        meteo_var : str, optional
            meteo variable can be "RH" or "RD". "RD" if you want data from a
            precipitation station (neerslagstation). "RH" if you want data from
            a meteo station. The default is "RH".
        stn : int, str or None, optional
            measurement station e.g. 829. The default is None.
        fname : str, path object, file-like object or None, optional
            filename of a knmi file. The default is None.
        xy : list, tuple or None, optional
            RD coördinates of a location in the Netherlands. The station nearest
            to this location used. The Default is None.
        start : str, datetime or None, optional
            start date of observations. The default is None.
        end : str, datetime or None, optional
            end date of observations. The default is None.
        fill_missing_obs : bool, optional
            if True nan values in time series are filled with nearby time series.
            The default is False.
        interval : str, optional
            desired time interval for observations. Options are 'daily' and
            'hourly'. The default is 'daily'.
        use_api : bool, optional
            if True the api is used to obtain the data, API documentation is here:
                https://www.knmi.nl/kennis-en-datacentrum/achtergrond/data-ophalen-vanuit-een-script
            if False a text file is downloaded into a temporary folder and the
            data is read from there. Default is True since the api is back
            online (July 2021).
        raise_exceptions : bool, optional
            if True you get errors when no data is returned. The default is False.

        Returns
        -------
        PrecipitationObs object with a precipitation time series and attributes
        """

        assert meteo_var in [
            "RD",
            "RH",
        ], 'meteo_var for precipitation should be "RD" or "RH"'

        return super().from_knmi(
            meteo_var,
            stn=stn,
            fname=fname,
            xy=xy,
            start=start,
            end=end,
            fill_missing_obs=fill_missing_obs,
            interval=interval,
            use_api=use_api,
            raise_exceptions=raise_exceptions,
            startdate=startdate,
            enddate=enddate,
        )

    @classmethod
    def from_wow(
        cls,
        stn: str = None,
        xy: List[float] = None,
        start: Optional[pd.Timestamp] = None,
        end: Optional[pd.Timestamp] = None,
    ):
        """Get a PrecipitationObs timeseries from a wow.knmi.nl station.

        Parameters
        ----------
        stn : Optional[int, str], optional
            station name
        xy : Optinal[List[float]], optinial
            longitude latitude of location [lon, lat] eg: [4.85, 51.95]
        start : Optional[pd.Timestamp], optional
            start date of observations, by default None
        end : Optional[pd.Timestamp], optional
            start date of observations, by default None

        Returns
        -------
        PrecipitationObs
        """
        return super().from_wow(
            meteo_var="rain_rate", stn=stn, xy=xy, start=start, end=end
        )



import logging
import os
import warnings

import numpy as np
import xarray as xr
from scipy.interpolate import griddata
from scipy.spatial import Delaunay

from ..observation import ModelObs

logger = logging.getLogger(__name__)


def read_imod_results(
    obs_collection,
    ml,
    runfile,
    mtime,
    model_ws,
    modelname="",
    nlay=None,
    exclude_layers=0,
):
    """Read imod model results at point locations.

    Parameters
    ----------
    obs_collection : ObsCollection
        collection of observations at which points imod results will be read
    ml : flopy.modflow.mf.model
        modflow model
    runfile : Runfile
        imod runfile object
    mtime : list of datetimes
        datetimes corresponding to the model periods
    model_ws : str
        model workspace with imod model
    nlay : int, optional
        number of layers if None the number of layers from ml is used.
    modelname : str
        modelname
    exclude_layers : int
        exclude modellayers from being read from imod
    """
    import imod

    if ml.modelgrid.xoffset == 0 or ml.modelgrid.yoffset == 0:
        warnings.warn(
            "you probably want to set the xll and/or yll attributes of ml.modelgrid"
        )

    if nlay is None:
        nlay = ml.modelgrid.nlay

    xmid, ymid, _ = ml.modelgrid.xyzcellcenters

    xy = np.array([xmid.ravel(), ymid.ravel()]).T
    uv = obs_collection.loc[:, ("x", "y")].dropna(how="any", axis=0).values
    vtx, wts = interp_weights(xy, uv)

    hm_ts = np.zeros((obs_collection.shape[0], len(mtime)))

    # loop over layers
    for m in range(nlay):
        if m < exclude_layers:
            continue
        mask = obs_collection.modellayer.values == m
        # loop over timesteps
        for t, date in enumerate(mtime):
            head_idf = "head_{}_l{}.idf".format(date.strftime("%Y%m%d"), m + 1)
            fname = os.path.join(
                model_ws, runfile.data["OUTPUTDIRECTORY"], "head", head_idf
            )

            logger.info(f"read {fname}")
            ihds, _attrs = imod.idf.read(fname)
            hm = interpolate(ihds, vtx, wts)
            hm_ts[mask, t] = hm[mask]

    mo_list = []
    for i, name in enumerate(obs_collection.index):
        mo = ModelObs(
            index=mtime,
            data=hm_ts[i],
            name=name,
            model=modelname,
            x=obs_collection.loc[name, "x"],
            y=obs_collection.loc[name, "y"],
            source="modflow",
            meta=obs_collection.loc[name, "obs"].meta,
        )
        mo_list.append(mo)

    return mo_list


def read_modflow_results(
    obs_collection,
    ml,
    hds_arr,
    mtime,
    modelname="",
    nlay=None,
    exclude_layers=None,
    method="linear",
):
    """Read modflow groundwater heads at points in obs_collection.

    Parameters
    ----------
    obs_collection : ObsCollection
        locations of model observation
    ml : flopy.modflow.mf.model
        modflow model
    hds_arr : numpy array
        heads with shape (ntimesteps, nlayers, nrow, ncol)
    mtime : list of datetimes
        dates for each model timestep
    modelname : str, optional
        modelname
    nlay : int, optional
        number of layers if None the number of layers from ml is used.
    exclude_layers : list of int, optional
        exclude the observations in these modellayers
    method : str, optional
        interpolation method, either 'linear' or 'nearest',
        default is linear.
    """
    if ml.modelgrid.grid_type == "structured":
        if ml.modelgrid.xoffset == 0 or ml.modelgrid.yoffset == 0:
            warnings.warn(
                "you probably want to set the xll and/or yll attributes in DIS!"
            )

    if isinstance(hds_arr, xr.DataArray):
        hds_arr = hds_arr.values

    if nlay is None:
        nlay = ml.modelgrid.nlay

    if modelname == "":
        modelname = ml.name

    xmid, ymid, _ = ml.modelgrid.xyzcellcenters

    if ml.modelgrid.grid_type == "structured":
        xy = np.array([xmid.ravel(), ymid.ravel()]).T
    elif ml.modelgrid.grid_type == "vertex":
        xy = np.array([xmid, ymid]).T

    uv = obs_collection.loc[:, ("x", "y")].dropna(how="any", axis=0).values
    if method == "linear":
        vtx, wts = interp_weights(xy, uv)

    # get interpolated timeseries from hds_arr
    hm_ts = np.nan * np.ones((obs_collection.shape[0], hds_arr.shape[0]))

    if exclude_layers is None:
        exclude_layers = []

    # loop over layers
    for m in range(nlay):
        if m in exclude_layers:
            continue
        mask = obs_collection["modellayer"].values == m
        # loop over timesteps
        for t in range(hds_arr.shape[0]):
            ihds = hds_arr[t, m]
            ihds[ihds <= -999.0] = np.nan
            if method == "linear":
                hm = interpolate(ihds, vtx, wts)
                hm_ts[mask, t] = hm[mask]
            elif method == "nearest":
                hm_ts[mask, t] = griddata(xy, ihds.ravel(), uv, method=method)[mask]
            else:
                raise ValueError(f"Unknown method: '{method}'")

    mo_list = []
    for i, name in enumerate(obs_collection.index):
        mo = ModelObs(
            index=mtime,
            data=hm_ts[i],
            name=name,
            model=modelname,
            x=obs_collection.loc[name, "x"],
            y=obs_collection.loc[name, "y"],
            meta=obs_collection.loc[name, "obs"].meta,
            source="modflow",
        )
        mo_list.append(mo)

    return mo_list


def interp_weights(xy, uv, d=2):
    """Calculate interpolation weights [1]_.

    Parameters
    ----------
    xy : np.array
        array containing x-coordinates in first column and y-coordinates
        in second column
    uv : np.array
        array containing coordinates at which interpolation weights should
        be calculated, x-data in first column and y-data in second column
    d : int, optional
        dimension of data? (the default is 2, which works for 2D data)

    Returns
    -------
    vertices: np.array
        array containing interpolation vertices
    weights: np.array
        array containing interpolation weights per point


    References
    ----------
    .. [1] https://stackoverflow.com/questions/20915502/
    speedup-scipy-griddata-for-multiple-interpolations-between-two-irregular-grids
    """

    tri = Delaunay(xy)
    simplex = tri.find_simplex(uv)
    vertices = np.take(tri.simplices, simplex, axis=0)
    temp = np.take(tri.transform, simplex, axis=0)
    delta = uv - temp[:, d]
    bary = np.einsum("njk,nk->nj", temp[:, :d, :], delta)
    return vertices, np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))


def interpolate(values, vtx, wts, fill_value=np.nan):
    """Interpolate values at locations defined by vertices and points [2]_, as
    calculated by interp_weights function.

    Parameters
    ----------
    values : np.array
        array containing values to interpolate
    vtx : np.array
        array containing interpolation vertices, see interp_weights()
    wts : np.array
        array containing interpolation weights, see interp_weights()
    fill_value : float
        fill value for points that have to be extrapolated (e.g. at or
        beyond edges of the known points)

    Returns
    -------
    arr: np.array
        array containing interpolated values at locations as given by
        vtx and wts

    References
    ----------
    .. [2] https://stackoverflow.com/questions/20915502/
    speedup-scipy-griddata-for-multiple-interpolations-between-two-irregular-grids
    """
    ret = np.einsum("nj,nj->n", np.take(values, vtx), wts)
    ret[np.any(wts < 0, axis=1)] = fill_value
    return ret



import datetime as dt
import os

import pandas as pd
import pastastore as pst
import pytest
from requests.exceptions import ConnectionError, HTTPError
from pathlib import Path

import hydropandas as hpd

datadir = Path(__file__).parent / "data"

# %% BRO


def test_bro_gld():
    # single observation
    bro_id = "GLD000000012893"
    hpd.GroundwaterObs.from_bro(bro_id)


def test_bro_gmn():
    # single observation
    bro_id = "GMN000000000001"  # 6 objects as per 2025-4-17
    hpd.read_bro(bro_id=bro_id, only_metadata=True)


def test_bro_extent():
    # extent = (210260, 213550, 459890, 473920)  # extent more than 1000 observations
    # extent = (213260, 213550, 473890, 473920)  # extent skip duplicates
    extent = (102395, 103121, 434331, 434750)  # 4 observations within extent

    hpd.read_bro(extent=extent, only_metadata=True)


def test_bro_extent_too_big():
    extent = (102395, 213550, 334331, 473920)  # too many observations in extent

    with pytest.raises(HTTPError):
        hpd.read_bro(extent=extent, only_metadata=True)


# %% DINO
dino2019 = datadir / "2019-Dino-test"
dino2024 = datadir / "2024-Dino-test"
dinozip = dino2019 / "dino.zip"


def test_observation_gwq():
    # single observation
    path = dino2019 / "Grondwatersamenstellingen_Put/B52C0057.txt"
    hpd.WaterQualityObs.from_dino(path)


def test_observation_wl():
    path = dino2019 / "Peilschaal/P58A0001.csv"
    hpd.WaterlvlObs.from_dino(path)


def observation_gw_dino_old():
    path = dino2019 / "Grondwaterstanden_Put/B33F0080001_1.csv"
    o = hpd.GroundwaterObs.from_dino(path=path)
    return o


def observation_gw_dino_new():
    path = dino2024 / "DINO_Grondwaterstanden/B02H0090001.csv"
    o = hpd.GroundwaterObs.from_dino(path=path)
    return o


def test_observation_gw():
    observation_gw_dino_old()
    observation_gw_dino_new()


def test_obscollection_from_directory_old_school():
    dino_gw = hpd.read_dino(
        dirname=dino2019 / "dino.zip",
        ObsClass=hpd.GroundwaterObs,
        subdir="Grondwaterstanden_Put",
        suffix="1.csv",
        keep_all_obs=True,
    )
    obs_list = [o for o in dino_gw.obs.values]
    hpd.ObsCollection(obs_list)


def test_obscollection_from_directory_new_school():
    hpd.read_dino(
        dirname=dino2024 / "dino.zip", ObsClass=hpd.GroundwaterObs, keep_all_obs=True
    )


def test_obscollection_from_df():
    df = pd.DataFrame(index=["pb1", "pb2"], data={"tube_nr": [1, 1]})

    hpd.ObsCollection(df)


def test_obscollection_empty():
    hpd.ObsCollection()


# read dino directories


def obscollection_dinozip_gw():
    # groundwater quantity
    oc = hpd.read_dino(
        dirname=dino2019 / "dino.zip",
        ObsClass=hpd.GroundwaterObs,
        subdir="Grondwaterstanden_Put",
        suffix="1.csv",
        keep_all_obs=False,
    )
    return oc


def obscollection_dinozip_gw_keep_all_obs():
    # do not delete empty dataframes
    oc = hpd.read_dino(
        dirname=dino2019 / "dino.zip",
        ObsClass=hpd.GroundwaterObs,
        subdir="Grondwaterstanden_Put",
        suffix="1.csv",
        keep_all_obs=True,
    )
    return oc


def obscollection_dinozip_wl():
    # surface water
    oc = hpd.read_dino(
        dirname=dino2019 / "dino.zip",
        ObsClass=hpd.WaterlvlObs,
        subdir="Peilschaal",
        suffix=".csv",
    )

    return oc


def test_obscollection_dinozip_gw():
    # groundwater quantity
    obscollection_dinozip_gw()


def test_obscollection_dinozip_gw_keep_all_obs():
    obscollection_dinozip_gw_keep_all_obs()


def test_obscollection_dinozip_wl():
    obscollection_dinozip_wl()


def test_obscollection_dinozip_gwq():
    # groundwater quality
    hpd.read_dino(
        dirname=dino2019 / "dino.zip",
        ObsClass=hpd.WaterQualityObs,
        subdir="Grondwatersamenstellingen_Put",
        suffix=".txt",
    )


# %% FEWS
fewsdir = datadir / "2019-FEWS-test"


def obscollection_fews_lowmemory():
    oc = hpd.read_fews(
        fewsdir / "WaalenBurg_201810-20190215_prod.zip",
        locations=None,
        low_memory=True,
    )
    return oc


def test_obscollection_fews_highmemory():
    hpd.read_fews(
        fewsdir / "WaalenBurg_201810-20190215_prod.zip",
        to_mnap=False,
        remove_nan=False,
        low_memory=False,
    )


def test_obscollection_fews_lowmemory():
    obscollection_fews_lowmemory()


def test_obscollection_fews_selection():
    hpd.read_fews(
        fewsdir / "WaalenBurg_201810-20190215_prod.zip",
        locations=("MPN-N-2",),
    )


# %% WISKI
wiskidir = datadir / "2019-WISKI-test"


@pytest.mark.slow
def test_observation_wiskicsv_gw():
    hpd.GroundwaterObs.from_wiski(
        wiskidir / "1016_PBF.csv",
        sep=r"\s+",
        header_sep=":",
        header_identifier=":",
        parse_dates={"datetime": [0, 1]},
        index_col=["datetime"],
        dayfirst=True,
        translate_dic={"name": "Station Number", "x": "GlobalX", "y": "GlobalY"},
    )


@pytest.mark.slow
def test_obscollection_wiskizip_gw():
    hpd.read_wiski(
        wiskidir / "1016_PBF.zip",
        translate_dic={"name": "Station Number", "x": "GlobalX", "y": "GlobalY"},
        sep=r"\s+",
        header_sep=":",
        dayfirst=True,
        header_identifier=":",
        parse_dates={"datetime": [0, 1]},
        index_col=["datetime"],
    )


# %% PASTASTORE
def test_to_pastastore():
    dino_gw = obscollection_dinozip_gw()
    # drop duplicate
    dino_gw.drop("B22D0155-001", inplace=True)
    pstore = dino_gw.to_pastastore()
    # export to zip for read test
    pstore.to_zip("test_pastastore.zip")


def test_from_pastastore():
    pstore = pst.PastaStore.from_zip(
        "test_pastastore.zip", conn=pst.DictConnector("pastas_db")
    )
    _ = hpd.read_pastastore(pstore, "oseries")
    os.remove("test_pastastore.zip")


# %% excel


def test_to_excel():
    oc = hpd.read_fews(
        fewsdir / "WaalenBurg_201810-20190215_prod.zip",
        locations=None,
        low_memory=True,
    )

    oc.to_excel("tests/data/excel/test.xlsx")


def test_from_excel():
    hpd.read_excel("tests/data/excel/test.xlsx")


# %% Meteo


def test_pressure_obs_from_stn():
    hpd.MeteoObs.from_knmi(
        stn=310, meteo_var="P", interval="hourly", start="2022-1-1", end="2023-1-1"
    )


def test_pressure_read_knmi():
    hpd.read_knmi(
        stns=(310,),
        meteo_vars=("P",),
        interval="hourly",
        starts="2022-1-1",
        ends="2023-1-1",
    )


# %% Evaporation

knmidir = datadir / "2023-KNMI-test"


def test_evap_obs_from_file():
    path = knmidir / "etmgeg_260.txt"
    hpd.EvaporationObs.from_knmi(fname=path)


def test_evap_obs_from_stn():
    hpd.EvaporationObs.from_knmi(stn=260, meteo_var="EV24")


def test_evap_obs_from_stn_makkink():
    hpd.EvaporationObs.from_knmi(stn=260, meteo_var="makkink")


def test_evap_obs_from_stn_penman():
    hpd.EvaporationObs.from_knmi(stn=260, meteo_var="penman")


def test_evap_obs_from_stn_hargreaves():
    hpd.EvaporationObs.from_knmi(stn=260, meteo_var="hargreaves")


# %% Precipitation


def test_precip_obs_from_file():
    path = knmidir / "neerslaggeg_ESBEEK_831.txt"
    hpd.PrecipitationObs.from_knmi(fname=path)


def test_precip_obs_from_stn():
    hpd.PrecipitationObs.from_knmi(stn=233, meteo_var="RD")


def test_knmi_obs_from_stn_no_api():
    hpd.PrecipitationObs.from_knmi(stn=233, meteo_var="RD", use_api=False)


def test_knmi_obs_from_stn_without_any_data():
    hpd.EvaporationObs.from_knmi(
        stn=210, start="19500101", end="19600101", fill_missing_obs=False
    )


def test_knmi_obs_from_stn_with_missing_data_in_time_period():
    hpd.PrecipitationObs.from_knmi(stn=441, meteo_var="RD", start="2010-1-2")


def test_knmi_obs_from_xy():
    hpd.PrecipitationObs.from_knmi(xy=(100000, 350000))


# @pytest.xfail(
#     "Station HEIBLOEM 967 not available. See issue"
#     " https://github.com/ArtesiaWater/hydropandas/issues/103"
# )
def test_knmi_collection_from_locations():
    obsc = obscollection_dinozip_gw()
    try:
        hpd.read_knmi(
            locations=obsc,
            meteo_vars=["EV24", "RD"],
            starts="2010",
            ends="2015",
            raise_exceptions=True,
        )
    except ConnectionError:
        pass


def test_knmi_collection_from_stns():
    stns = [344, 260]  # Rotterdam en de Bilt
    hpd.read_knmi(
        stns=stns,
        meteo_vars=["EV24", "RH"],
        starts=["2010", "2010"],
        ends=["2015", "2015"],
    )


def test_knmi_collection_from_grid():
    # somewhere in Noord-Holland (near Castricum)
    xy = [[104150.0, 510150.0], [104550.0, 510550.0]]
    hpd.read_knmi(
        xy=xy,
        meteo_vars=["RH"],
        starts=["2010"],
        ends=["2015"],
    )


# %% WATERINFO

waterinfodir = datadir / "2023-waterinfo-test"


def test_waterinfo_from_dir():
    path = waterinfodir
    hpd.read_waterinfo(file_or_dir=path)


def test_waterinfo_ddlpy():
    grootheid_code = "WATHTE"
    locatie = "SCHOONHVN"
    tmin = dt.datetime(2020, 1, 1)
    tmax = dt.datetime(2020, 1, 5)
    hpd.WaterlvlObs.from_waterinfo(
        grootheid_code=grootheid_code, locatie=locatie, tmin=tmin, tmax=tmax
    )


def test_waterinfo_ddlpy_extent():
    grootheid_code = "WATHTE"
    tmin = dt.datetime(2020, 1, 1)
    tmax = dt.datetime(2020, 1, 2)
    extent = (110000, 125000, 429550, 449900)
    oc = hpd.read_waterinfo(
        extent=extent, grootheid_code=grootheid_code, tmin=tmin, tmax=tmax
    )
    assert not oc.empty


# %% MENYANTHES


def test_obscollection_menyanthes():
    fname = datadir / "2023-MEN-test" / "test.men"
    hpd.read_menyanthes(fname, ObsClass=hpd.GroundwaterObs)



import logging

from hydropandas.io import dino
from pathlib import Path

logging.basicConfig(level=logging.DEBUG)

datadir = Path(__file__).parent / "data"
dino2019 = datadir / "2019-Dino-test"
dino2024 = datadir / "2024-Dino-test"


def test_dino_csv_old_style():
    fname = dino2019 / "Grondwaterstanden_Put/B37A0112001_1.csv"
    dino.read_dino_groundwater_csv(fname)


def test_dino_csv_new_style():
    fname = dino2024 / "DINO_Grondwaterstanden" / "B02H0089001.csv"
    dino.read_dino_groundwater_csv(fname)


def test_dino_csv_duplicate_index():
    # contains 1 duplicate index 2019-11-19
    fname = dino2019 / "Grondwaterstanden_Put" / "B22D0155001_1.csv"
    measurements, _ = dino.read_dino_groundwater_csv(fname)

    # check if measurements contains duplicate indices
    assert measurements.index.duplicated().any()

    measurements, _ = dino.read_dino_groundwater_csv(
        fname, remove_duplicates=True, keep_dup="last"
    )

    # check if measurements contains no duplicate indices
    assert ~measurements.index.duplicated().any()



import pandas as pd
import pytest

import hydropandas as hpd
from hydropandas.io import wow


def test_wow_strftime() -> None:
    timestamp = pd.Timestamp(year=2010, month=10, day=2, hour=10, minute=34, second=0)
    timestampstr = wow._wow_strftime(timestamp)
    assert timestampstr == "2010-10-02T10%3A30%3A00"


@pytest.mark.skip
def test_get_wow_stn() -> None:
    stn = "423216079"  # Macquarie Island
    start = pd.Timestamp(year=2023, month=6, day=4, hour=1, minute=34, second=0)
    end = pd.Timestamp(year=2023, month=6, day=5, hour=23, minute=54, second=0)
    obs = hpd.MeteoObs.from_wow(stn=stn, meteo_var="temperature", start=start, end=end)
    assert obs.name == "Macquarie Island"


@pytest.mark.skip
def test_get_wow_nearest() -> None:
    lat = -35.4184
    lon = 149.0937
    xy = [lon, lat]
    start = pd.Timestamp(year=2023, month=6, day=4, hour=1, minute=34, second=0)
    end = pd.Timestamp(year=2023, month=6, day=5, hour=23, minute=54, second=0)
    obs = hpd.PrecipitationObs.from_wow(xy=xy, start=start, end=end)
    assert obs.name == "Hillanhome"



# flake8: noqa
import logging

from .extensions import geo as _geo
from .extensions import gwobs as _gwobs
from .extensions import plots as _plots
from .extensions import stats as _stats
from .io.fews import get_fews_pid
from .obs_collection import (
    ObsCollection,
    read_bro,
    read_bronhouderportaal_bro,
    read_dino,
    read_excel,
    read_fews,
    read_imod,
    read_knmi,
    read_lizard,
    read_matroos,
    read_menyanthes,
    read_modflow,
    read_pastastore,
    read_pickle,
    read_waterconnect,
    read_waterinfo,
    read_wiski,
)
from .observation import (
    EvaporationObs,
    GroundwaterObs,
    MeteoObs,
    ModelObs,
    Obs,
    PrecipitationObs,
    WaterlvlObs,
    WaterQualityObs,
)
from .rcparams import rcParams
from .version import __version__, show_versions

logging.getLogger("hydropandas").addHandler(logging.NullHandler())



import pytest

import hydropandas as hpd


# Original tests when only Vitens was supported
def test_single_observation():
    code = "27BP0003"
    o = hpd.GroundwaterObs.from_lizard(code)
    assert o.tube_nr == 1


def test_extent():
    extent = [201500, 202000, 502000, 502200]
    oc = hpd.read_lizard(extent)
    assert not oc.empty
    assert not oc.get_obs(name=oc.index[0]).empty


@pytest.mark.slow
def test_codes():
    oc = hpd.read_lizard(
        codes=["39F-0735", "39F-0736", "39F-0737"], combine_method="merge"
    )
    assert not oc.empty


@pytest.mark.slow
def test_many_tubed_well():
    oc = hpd.read_lizard(codes="EEWP004", tube_nr="all")
    assert not oc.empty


@pytest.mark.slow
def test_complex_well():
    oc = hpd.read_lizard(codes="BUWP014", tube_nr="all")
    assert not oc.empty


def test_combine():
    hpd.GroundwaterObs.from_lizard("39F-0736", tube_nr=1, combine_method="columns")
